"""This is the demo script to show how administrator can manage security
filters.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.access_and_security.security_filter import (
    AttributeFormSystemRef,
    AttributeRef,
    ConstantParameter,
    ConstantType,
    list_security_filters,
    LogicFunction,
    LogicOperator,
    PredicateForm,
    PredicateFormFunction,
    Qualification,
    SecurityFilter
)
from mstrio.object_management import full_search
from mstrio.types import ObjectTypes
from mstrio.users_and_groups import list_user_groups, list_users

from mstrio.connection import get_connection

PROJECT_NAME = '<project_name>'  # Project to connect to

SECURITY_FILTER_ID = '<sec_filter_id>'
SECURITY_FILTER_NAME = '<sec_filter_name>'
SECURITY_FILTER_NAME_NEW = '<new_sec_filter_name>'
SECURITY_FILTER_NAME_2 = '<sec_filter_name_2>'
SECURITY_FILTER_DESCRIPTION = '<sec_filter_description>'
QUALIFICATION = Qualification(
    tree=PredicateForm(
        # see access_and_security/security_filter/predicates.py
        # for PredicateFormFunction values
        function=PredicateFormFunction.GREATER,
        # see access_and_security/security_filter/predicate_parameters.py
        # for ConstantType values
        parameters=[ConstantParameter(ConstantType.DOUBLE, '<value>')],
        # insert object id and its attribute name to which you want to refer to
        attribute=AttributeRef('<object_id>', '<attribute_name>'),
        # insert object id and its attribute form system name
        # to which you want to refer to
        form=AttributeFormSystemRef('<object_id>', '<attribute_form_system_name>')
    )
)
CUSTOMER_ATTRIBUTE_ID = '<object_with_attribute_id>'  # id of an object that uses the attribute
PROJECT_ID = '<project_id>'  # project in which a security filter will be searched for
FOLDER_ID = '<folder_id>'  # folder in which new security filter will be created
FOLDER_ID_NEW = '<new_folder_id>'  # folder to which a security filter will be moved

# see ObjectTypes enum in types.py for available values
OBJECT_TYPES = ObjectTypes.ATTRIBUTE_FORM
USED_BY_OBJECT_TYPE = ObjectTypes.ATTRIBUTE

# see access_and_security/security_filter/predicates.py
# for PredicateFormFunction values
PREDICATE_FORM_FUNCTION = PredicateFormFunction.EQUALS
# see access_and_security/security_filter/predicate_parameters.py
# for ConstantType values
PREDICATE_FORM_PARAM = ConstantParameter(ConstantType.STRING, '<value>')
# insert object id and its attribute name to which you want to refer to
PREDICATE_FORM_ATTRIBUTE = AttributeRef('<object_id>', '<attribute_name>')
# insert object id and its attribute form system name
# to which you want to refer to
PREDICATE_FORM_ATTRIBUTE_FORM = AttributeFormSystemRef(
    '<object_id>', '<attribute_form_system_name>'
)
# insert data locale used to select value of the form
PREDICATE_FORM_DATA_LOCALE = '<locale>'

# see access_and_security/security_filter/predicates.py
# for PredicateFormFunction values
PREDICATE_FORM_FUNCTION_2 = PredicateFormFunction.EQUALS
# see access_and_security/security_filter/predicate_parameters.py
# for ConstantType values
PREDICATE_FORM_PARAM_2 = ConstantParameter(ConstantType.STRING, '<value>')
# insert object id and its attribute name to which you want to refer to
PREDICATE_FORM_ATTRIBUTE_2 = AttributeRef('<object_id>', '<attribute_name>')
# insert object id and its attribute form system name
# to which you want to refer to
PREDICATE_FORM_ATTRIBUTE_FORM_2 = AttributeFormSystemRef(
    '<object_id>', '<attribute_form_system_name>'
)
# insert data locale used to select value of the form
PREDICATE_FORM_DATA_LOCALE_2 = '<locale>'

PARAMETER_CHANGED_VALUE = '<value>'  # value of paramater to change to
USER_NAME_BEGINS = '<username_begins>'  # beginning of username to look for
USER_GROUP_NAME_BEGINS = '<usergroup_name_begins>'  # beginning of usergroup to look for

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# list all security filters
list_security_filters(conn)

# initialize security filter by its id
sec_fil = SecurityFilter(conn, id=SECURITY_FILTER_ID)

# show information, qualification and members of the security filter
sec_fil.information.to_dict()
sec_fil.qualification.to_dict()
sec_fil.members

# create new security filter
name = SECURITY_FILTER_NAME
folder_id = FOLDER_ID
qualification = QUALIFICATION

new_sec_fil = SecurityFilter.create(conn, qualification, name, folder_id)

# Create new security filter with more than one attribute form.

# 1. Find attribute forms for the particular attribute
# You have to look for objects with type ObjectTypes.ATTRIBUTE_FORM which are
# used by object with type ObjectTypes.ATTRIBUTE and ID of particular attribute.
attr_forms_of_customer_attr = full_search(
    conn,
    project_id=PROJECT_ID,
    object_types=OBJECT_TYPES,
    to_dictionary=False,
    used_by_object_id=CUSTOMER_ATTRIBUTE_ID,
    used_by_object_type=USED_BY_OBJECT_TYPE,
)
# just a simple example of extracting just `name` and `id` from each dict of
# attribute form which was returned
tmp_dict = [{'id': a.id, 'name': a.name} for a in attr_forms_of_customer_attr]

# 2. To build security filter qualification where there are two attribute
# forms you have to create two separate `PredicateForm`s and connect them with
# `LogicOperator`. You cannot set values for two attribute forms in a single
# `PredicateForm`.
pf1 = PredicateForm(
    function=PREDICATE_FORM_FUNCTION,
    parameters=[PREDICATE_FORM_PARAM],
    attribute=PREDICATE_FORM_ATTRIBUTE,
    form=PREDICATE_FORM_ATTRIBUTE_FORM,
    data_locale=PREDICATE_FORM_DATA_LOCALE,
)
pf2 = PredicateForm(
    function=PREDICATE_FORM_FUNCTION_2,
    parameters=[PREDICATE_FORM_PARAM_2],
    attribute=PREDICATE_FORM_ATTRIBUTE_2,
    form=PREDICATE_FORM_ATTRIBUTE_FORM_2,
    data_locale=PREDICATE_FORM_DATA_LOCALE_2,
)
# see access_and_security/security_filter/predicates.py for LogicFunction values
logic_operator = LogicOperator(function=LogicFunction.AND, children=[pf1, pf2])
qualification2 = Qualification(tree=logic_operator)
name2 = SECURITY_FILTER_NAME_2
new_sec_fil2 = SecurityFilter.create(conn, qualification2, name2, folder_id)

# alter security filter - its name, description, qualification and folder
new_name = SECURITY_FILTER_NAME_NEW
new_description = SECURITY_FILTER_DESCRIPTION
# copy old qualification and change year
new_qualification = Qualification.from_dict(new_sec_fil.qualification.to_dict())
new_qualification.tree.parameters[0].constant['value'] = PARAMETER_CHANGED_VALUE
new_folder_id = FOLDER_ID_NEW
sec_fil.alter(qualification=new_qualification, name=new_name, description=new_description)

# prepare users and user groups for applying and revoking security filter
mstr_user = list_users(conn, name_begins=USER_NAME_BEGINS)[0]
system_monitors = list_user_groups(conn, name_begins=USER_GROUP_NAME_BEGINS)[0]

# apply user(s) and/or user group(s) to security filter
sec_fil.apply([mstr_user, system_monitors.id])

# revoke user(s) and/or user group(s) from security filter
sec_fil.revoke([mstr_user.id, system_monitors])
sec_fil.revoke(sec_fil.members)  # revoke all members of security filter

# apply security filter directly to user or user group
mstr_user.apply_security_filter(sec_fil)

# revoke security filter directly from user or user group
system_monitors.apply_security_filter(sec_fil.id)

# fetch the latest security filter state
sec_fil.fetch(fetch_definition=False, fetch_members=True)

# delete security filter. When `force` argument is set
# to `False` (default value) then deletion must be approved.
sec_fil.delete(force=True)



from collections import defaultdict
from enum import auto
import logging
from typing import List, Optional, TYPE_CHECKING, Union

from mstrio import config
from mstrio.api import devices, objects
from mstrio.distribution_services.device.device_properties import (
    AndroidDeviceProperties,
    EmailDeviceProperties,
    FileDeviceProperties,
    FtpDeviceProperties,
    IOSDeviceProperties,
    PrinterDeviceProperties
)
from mstrio.distribution_services.transmitter import Transmitter
from mstrio.types import ObjectTypes
from mstrio.users_and_groups import User
from mstrio.utils.entity import DeleteMixin, Entity
from mstrio.utils.enum_helper import AutoName, get_enum_val
from mstrio.utils.helper import (
    delete_none_values,
    Dictable,
    fetch_objects,
    get_args_from_func,
    get_default_args_from_func,
    get_objects_id
)
from mstrio.utils.version_helper import class_version_handler

if TYPE_CHECKING:
    from mstrio.connection import Connection

logger = logging.getLogger(__name__)


class DeviceType(AutoName):
    EMAIL = auto()
    FTP = auto()
    FILE = auto()
    PRINTER = auto()
    IPHONE = auto()
    IPAD = auto()
    ANDROID = auto()
    ALL = auto()
    UNSUPPORTED = auto()


def list_devices(
    connection: "Connection", to_dictionary: bool = False, limit: int = None, **filters
) -> Union[List["Device"], List[dict]]:
    """Get list of Device objects or dicts. Optionally filter the
    devices by specifying filters.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        to_dictionary: If True returns dict, by default (False) returns
            Device objects.
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'description',
            'date_created', 'date_modified', 'acg']

    Examples:
        >>> list_devices(connection, name='device_name')
    """
    return Device._list_devices(
        connection=connection,
        to_dictionary=to_dictionary,
        limit=limit,
        **filters,
    )


@class_version_handler('11.3.0100')
class Device(Entity, DeleteMixin):
    """Devices are Distribution Services components that specify the format
     and transmission process of subscribed reports and documents.
     They are instances of transmitters that contain specific settings
     specific to a user’s environments.

    Attributes:
        name: name of the device
        id: identifier of the Device
        description: description of the device
        device_type: type of the Device, DeviceType Enum
        transmitter: information of Transmitter attached
        device_properties: properties of the device,
            different for every device type
        ancestors: List of ancestor folders
        type: Object type. Enum
        subtype: Object subtype
        ext_type: Object extended type.
        date_created: Creation time, DateTime object
        date_modified: Last modification time, DateTime object
        owner: User object that is the owner
        version: Object version ID
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
    """
    _DELETE_NONE_VALUES_RECURSION = False
    _OBJECT_TYPE = ObjectTypes.SUBSCRIPTION_DEVICE
    _DEVICE_TYPE_MAP = {
        "android": AndroidDeviceProperties,
        "email": EmailDeviceProperties,
        "file": FileDeviceProperties,
        "ftp": FtpDeviceProperties,
        "ipad": IOSDeviceProperties,
        "iphone": IOSDeviceProperties,
        "printer": PrinterDeviceProperties
    }
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'device_type': DeviceType,
        'owner': User.from_dict,
        'transmitter': Transmitter.from_dict,
        'device_properties': lambda source,
        connection,
        device_type_map=_DEVICE_TYPE_MAP: device_type_map[next(iter(source))].
        from_dict(source[next(iter(source))], connection)
    }
    _API_GETTERS = {
        (
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl'
        ): objects.get_object_info,
        ('id', 'name', 'description', 'device_type', 'transmitter',
         'device_properties'): devices.get_device
    }
    _API_DELETE = staticmethod(devices.delete_device)
    _API_PATCH: dict = {
        ("name", "description", "device_properties"): (devices.update_device, "put")
    }
    _PATCH_PATH_TYPES = {"name": str, "description": str, "device_properties": dict}

    def __init__(self, connection: "Connection", name: str = None, id: str = None):
        """Initialize Device object."""

        if id is None and name is None:
            raise ValueError("Please specify either 'id' or 'name' parameter in the constructor.")

        if id is None:
            objects_info = Device._list_devices(
                connection=connection,
                name=name,
                to_dictionary=True,
            )
            if objects_info:
                object_info, object_info["connection"] = objects_info[0], connection
                self._init_variables(**object_info)
            else:
                raise ValueError(f"There is no Device: '{name}'")
        else:
            super().__init__(connection=connection, object_id=id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        device_type = kwargs.get("device_type")
        self._device_type = DeviceType(device_type) if device_type else None
        self._transmitter = Transmitter.from_dict(kwargs.get("transmitter"), self.connection
                                                  ) if kwargs.get("transmitter") else None
        device_properties = kwargs.get("device_properties")
        self.device_properties = self._DEVICE_TYPE_MAP[device_type].from_dict(
            device_properties[device_type],
            self.connection,
        ) if device_properties and device_type else None

    @classmethod
    def create(
        cls,
        connection: "Connection",
        name: str,
        device_type: Union[DeviceType, str],
        transmitter: Union[Transmitter, str],
        device_properties: Union[dict, Dictable],
        description: str = None
    ) -> "Device":
        """Create a new device.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`
            name: device object name
            device_type: type of the device
            transmitter: Transmitter object
            description: device object description
            device_properties: properties of the device
        Returns:
            Device object.
        """
        device_type = get_enum_val(device_type, DeviceType)
        device_properties = device_properties.to_dict(
        ) if isinstance(device_properties, Dictable) else device_properties
        transmitter_id = get_objects_id(transmitter, Transmitter)
        body = {
            "name": name,
            "description": description,
            "deviceType": device_type,
            "transmitter": {
                "id": transmitter_id,
            },
            "deviceProperties": {
                device_type: device_properties
            }
        }
        body = delete_none_values(body, recursion=True)
        response = devices.create_device(connection, body).json()
        if config.verbose:
            logger.info(
                f"Successfully created device named: '{name}' with ID: '{response['id']}'."
            )
        return cls.from_dict(source=response, connection=connection)

    def alter(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        device_properties: Optional[Union[Dictable, dict]] = None
    ):
        """Alter the device object properties.

        Args:
            name: device object name
            description: device object description
            device_properties: properties of the device, specific for each
                device type
        """
        device_properties = self.device_properties if not device_properties else device_properties
        device_properties = device_properties.to_dict(
        ) if isinstance(device_properties, Dictable) else device_properties
        func = self.alter
        args = get_args_from_func(func)
        defaults = get_default_args_from_func(func)
        defaults_dict = dict(zip(args[-len(defaults):], defaults)) if defaults else {}
        local = locals()
        properties = defaultdict(dict)
        for property_key in defaults_dict.keys():
            if property_key == 'device_properties':
                properties[property_key][self.device_type.value] = device_properties
            elif local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    def update_properties(self) -> None:
        """Save compatible local changes of the object attributes to the
        I-Server.

        Raises:
            requests.HTTPError: if I-Server raises exception
        """
        changes = {k: v[1] for k, v in self._altered_properties.items()}
        if "device_properties" in changes:
            device_properties = changes["device_properties"]
            del changes["device_properties"]
        else:
            device_properties = self.device_properties
        changes["device_properties"] = {}
        changes["device_properties"][self.device_type.value] = device_properties.to_dict(
        ) if not isinstance(device_properties, dict) else device_properties
        self._alter_properties(**changes)
        self._altered_properties.clear()

    @classmethod
    def _list_devices(
        cls, connection: "Connection", to_dictionary: bool = False, limit: int = None, **filters
    ) -> Union[List["Device"], List[dict]]:
        objects = fetch_objects(
            connection=connection,
            api=devices.get_devices,
            dict_unpack_value="devices",
            limit=limit,
            filters=filters,
        )
        if to_dictionary:
            return objects
        return [cls.from_dict(source=obj, connection=connection) for obj in objects]

    @property
    def device_type(self):
        return self._device_type

    @property
    def transmitter(self):
        return self._transmitter



# flake8: noqa
from typing import Union

from .user import create_users_from_csv, list_users, User
from .user_connections import UserConnections
from .user_group import list_user_groups, UserGroup

UserOrGroup = Union[str, User, UserGroup]



"""This is the sample script to show how to manage filters.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to ease
its usage.
"""

from mstrio.connection import Connection, get_connection
from mstrio.modeling import (
    ExpressionFormat, Filter, list_filters, SchemaManagement, SchemaUpdateType
)

# Following variables are defining basic filters
PROJECT_NAME = "<Project_name>"  # Insert name of project here
SEARCH_NAME = "<Search_name>"  # Insert part of the filter name to be searched for
FILTER_ID1 = "<Filter_ID>"  # Insert ID of existing filter here
FILTER_ID2 = "<Filter_ID>"  # Insert ID of existing filter here
FILTER_ID3 = "<Filter_ID>"  # Insert ID of existing filter here
FILTER_NAME = "<Filter_name>"  # Insert name of a new filter to be created
FILTER_NEW_NAME = "<Filter_name_alter>"  # Insert new name of edited Filter here
FILTER_NEW_DESC1 = "<Filter_desc>"  # Insert new description of edited filter here
FILTER_NEW_DESC2 = "<Filter_desc>"  # Insert new description of edited filter here
FOLDER_ID = "<Folder_ID>"  # Insert folder ID here

conn: Connection = get_connection(workstationData, PROJECT_NAME)

# Example filter qualification data with expression specified as tree
FILTER_QUALIFICATION_AS_TREE = {
    "tree": {
        "type": "predicate_metric_qualification",
        "predicateTree": {
            "function": "greater",
            "parameters": [
                {
                    "parameterType": "constant",
                    "constant": {
                        "type": "int64", "value": "1"
                    },
                }
            ],
            "levelType": "none",
            "metric": {
                "objectId": "CE8DAE7048998FFECF77418BDF2343B8",
                "subType": "metric",
                "name": "Item Count",
            },
            "metricFunction": "value",
            "isIndependent": 0,
            "nullInclude": 0,
        },
    },
}

# Example filter qualification data with expression specified as tokens
FILTER_QUALIFICATION_AS_TOKENS = {
    "tokens": [
        {
            "level": "resolved",
            "state": "initial",
            "value": "%",
            "type": "character",
        },
        {
            "level": "resolved",
            "state": "initial",
            "value": "[Item Count]",
            "type": "object_reference",
        },
        {
            "level": "resolved",
            "state": "initial",
            "value": ">",
            "type": "character",
        },
        {
            "level": "resolved", "state": "initial", "value": "1", "type": "unknown"
        },
        {
            "level": "resolved",
            "state": "initial",
            "value": "",
            "type": "end_of_text",
        },
    ],
}

# Filters management
# Get list of filters, with examples of different conditions
list_of_all_filters = list_filters(connection=conn)
list_of_limited_filters = list_filters(connection=conn, limit=10)
list_of_limited_filters_to_dict = list_filters(connection=conn, limit=10, to_dictionary=True)
list_of_filters_by_name = list_filters(connection=conn, name=SEARCH_NAME)

# list of filters with expressions represented as tree
list_of_filters_as_trees = list_filters(connection=conn, show_expression_as=ExpressionFormat.TREE)
# list of filters with expressions represented as tokens
list_of_filters_as_tokens = list_filters(
    connection=conn, show_expression_as=ExpressionFormat.TOKENS
)

# list of filters with qualification represented also in tokens format
list_of_filters_with_qualification_as_tokens = list_filters(
    connection=conn, show_filter_tokens=True
)

# create new empty filter
test_filter = Filter.create(
    connection=conn,
    name=FILTER_NAME,
    destination_folder=FOLDER_ID,
)

# create new filter using qualification as tree and
# specify to return qualification also in tokens format
test_filter_tree = Filter.create(
    connection=conn,
    name=(FILTER_NAME + "_tree"),
    destination_folder=FOLDER_ID,
    show_filter_tokens=True,
    qualification=FILTER_QUALIFICATION_AS_TREE,
)

# create new filter using qualification as tokens
test_filter_tokens = Filter.create(
    connection=conn,
    name=(FILTER_NAME + "_tokens"),
    destination_folder=FOLDER_ID,
    qualification=FILTER_QUALIFICATION_AS_TOKENS,
)

# Any changes to schema object must be followed by schema_reload
# in order to use them in reports, dossiers and so on
schema_manager = SchemaManagement(connection=conn, project_id=conn.project_id)
task = schema_manager.reload(update_types=[SchemaUpdateType.LOGICAL_SIZE])

# Alter filter
test_filter.alter(name=FILTER_NEW_NAME, description=FILTER_NEW_DESC1)

# Alter filter description
test_filter_tokens.alter(description=FILTER_NEW_DESC2)

# Delete newly created filter
test_filter.delete(force=True)

# Get specific filter by id with expressions represented as trees
filter_ = Filter(connection=conn, id=FILTER_ID1)

# Get specified filter by id with expressions represented as tokens
filter_tokens_expression = Filter(
    connection=conn, id=FILTER_ID2, show_expression_as=ExpressionFormat.TOKENS
)

# Get specified filter by id with qualification also in tokens format
filter_tokens_qualification = Filter(connection=conn, id=FILTER_ID2, show_filter_tokens=True)

# List filter properties
properties = filter_.list_properties()



from functools import cached_property
import logging
from typing import Optional, Type

from mstrio.api import datasources
from mstrio.api import tables as tables_api
from mstrio.connection import Connection
from mstrio.datasources.datasource_instance import DatasourceInstance
from mstrio.modeling.schema import ObjectSubType, SchemaObjectReference
from mstrio.modeling.schema.helpers import TableColumn, TableColumnMergeOption
from mstrio.modeling.schema.table.logical_table import list_logical_tables, LogicalTable
from mstrio.utils.helper import Dictable, fetch_objects
from mstrio.utils.version_helper import method_version_handler

logging.getLogger().setLevel(logging.INFO)


@method_version_handler('11.3.0100')
def list_datasource_warehouse_tables(
    connection: Connection,
    datasource_id: str,
    namespace_id: str,
    name: str = None,
    to_dictionary: bool = False,
    limit: Optional[int] = None
) -> list[Type["WarehouseTable"]] | list[dict]:
    """Lists available warehouse tables in a specified datasource within a
       specified namespace.

    Args:
        connection (Connection): Object representation of MSTR Connection.
        datasource_id (str): ID of a datasource.
        namespace_id (str): ID of a namespace within a given datasource_id.
        name (str): string by which to filter the table name
        to_dictionary (bool, optional): If True, the function will return list
            of dictionaries. Otherwise, list of WarehouseTable objects.
            Defaults to False.
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.

    Returns:
        Union[list["WarehouseTable"], list[dict]]: A list of WarehouseTable
            or dictionary objects representing warehouse tables in a specified
            namespace in a specified datasource.

    Examples:
        >>> datasource_id = "DAD6CAD6457DAF29E34463961688EA60"
        >>> namespace_id = "eyJucyI6InB1YmxpYyJ9"
        >>> available_wh_tables = list_datasource_warehouse_tables(
            connection, datasource_id, namespace_id)
    """

    tables = fetch_objects(
        connection=connection,
        api=tables_api.get_available_warehouse_tables,
        limit=limit,
        filters={'name': name} if name else None,
        dict_unpack_value="tables",
        datasource_id=datasource_id,
        namespace_id=namespace_id,
    )
    [
        table.update({
            "datasource": {
                "id": datasource_id
            }, "namespace_id": namespace_id
        }) for table in tables
    ]
    if to_dictionary:
        return tables
    return WarehouseTable.bulk_from_dict(source_list=tables, connection=connection)


class WarehouseTable(Dictable):
    """An object representation of a warehouse table. Depending on whether a
    warehouse table is included in a project, it may or may not have a
    persistent id and associated PhysicalTable object. When you add a warehouse
    table to a project, a MSTR server will create a PhysicalTable object that
    represents the added warehouse table and will assign it a persistent ID.
    You can then manipulate the table further using functionalities in
    mstrio.modeling.schema.table.physical_table.

    Attributes:
        connection (Connection): An object representation of MSTR Connection.
        datasource (DatasourceInstance): An object representation of datasource
            instance in which the table resides.
        name (str): A name of the table.
        namespace (str): A name of a namespace in which the table resides.
        namespace_id (str): An ID of a namespace in which the table resides.
        physical_table_id(str): An ID of an associated PhysicalTable object.
            This attribute is set automatically after `add_to_project()` method
            was called on the table.
        dependent_logical_tables (Union[list[dict], list[LogicalTable]]): A list
            of logical tables that are mapped to the warehouse table.
        columns (Union[list[dict], list[TableColumn]]): A list of table columns.

    """

    _FROM_DICT_MAP: dict = {"datasource": DatasourceInstance.from_dict}
    _DELETE_NONE_VALUES_RECURSION = False
    available_tables: list["WarehouseTable"] = []

    def __init__(
        self,
        id: str,
        connection: Connection,
        datasource: DatasourceInstance,
        namespace_id: str,
        name: str,
        namespace: str,
    ):
        self.__id = id  # Non-persistent. It will change if the table name or namespace changes.
        self.connection = connection
        self.datasource = datasource
        self.name = name
        self.namespace = namespace
        self.namespace_id = namespace_id
        # When table is added to a project, it will get a persistent ID
        self.physical_table_id = None
        self._dependent_logical_tables = None
        self._columns = None

    def add_to_project(
        self,
        logical_table_name: str,
        prefix: Optional[str] = None,
        col_merge_option: Optional[TableColumnMergeOption | str] = None,
    ) -> Type["LogicalTable"]:
        """Adds a table to a project. This function corresponds to 'ADD WHTABLE'
           statement from MSTR Command Manager.

        Args:
            logical_table_name (str): Name of the logical table which will be
                mapped to this warehouse table.
            prefix (Optional[str], optional): Table prefix. Defaults to None.
            col_merge_option (Optional[Union[TableColumnMergeOption, str]]:
                Defines a column merge option. Defaults to None.

        Returns:
            LogicalTable: A LogicalTable object representing a logical table
                mapped to this warehouse table.
        """
        lt = LogicalTable.create(
            connection=self.connection,
            table_name=logical_table_name,
            primary_data_source=SchemaObjectReference(
                object_id=self.datasource.id, sub_type=ObjectSubType.LOGICAL_TABLE
            ),
            column_merge_option=col_merge_option,
            physical_table_prefix=prefix,
            physical_table_name=self.name,
            physical_table_namespace=self.namespace,
        )
        self.physical_table_id = lt.physical_table.id
        self.list_dependent_logical_tables(refresh=True)
        return lt

    def list_dependent_logical_tables(self, to_dictionary: bool = False,
                                      refresh: bool = False) -> list["LogicalTable"] | list[dict]:
        """Get all dependent logical tables.

        Args:
            to_dictionary (bool, optional): If True, returns a list of
                dictionaries. Defaults to False.
            refresh(bool, optional): If True, refreshes the result stored in
                `self.dependent_logical_tables`.

        Returns:
            Union[list[LogicalTable], list[dict]]: A list of LogicalTable
                objects or dictionaries representing logical tables.
        """
        if self._dependent_logical_tables and not refresh:
            return self._dependent_logical_tables
        logical_tables = list_logical_tables(connection=self.connection)
        if self.physical_table_id:
            dependent_logical_tables = [
                table for table in logical_tables
                if table.physical_table.id == self.physical_table_id
            ]
        else:
            dependent_logical_tables = [
                table for table in logical_tables if table.physical_table.table_name == self.name
            ]
        if refresh:
            self._dependent_logical_tables = dependent_logical_tables
        if to_dictionary:
            return [table.to_dict() for table in dependent_logical_tables]
        return dependent_logical_tables

    def delete_from_project(self, force: bool = False) -> Optional[list[bool]]:
        """Remove the Warehouse Table from a project by removing all of its
        dependent logical tables. It will only delete dependent logical
        tables if they themselves have no dependent objects. This function
        corresponds to "REMOVE WHTABLE" statement from MSTR Command Manager.
        """
        dependent_logical_tables = self.list_dependent_logical_tables(refresh=True)
        if dependent_logical_tables:
            logging.warning("Following logical tables will be deleted: ")
            [logging.info(f"{str(table)}") for table in dependent_logical_tables]
            confirmed_delete = (force or input("Would you like to continue? Y/n").lower() == "y")
            if confirmed_delete:
                statuses = [table.delete(force=True) for table in dependent_logical_tables]
                self.physical_table_id = (None if all(statuses) else self.physical_table_id)
                return statuses
        else:
            logging.error("This table is not included in a project.")

    def list_columns(self, to_dictionary: bool = False,
                     refresh: bool = False) -> list[Type[TableColumn]] | list[dict]:
        """Get columns for a specific database table.

        Args:
            to_dictionary (bool, optional): If True, returns a list of
                dictionaries. Defaults to False.
            refresh (bool, optional): If True, refetches and overwrites the
                cached version of results stored in
                WarehouseTable.available_tables.

        Returns:
            Union[list[TableColumn], list[dict]]: A list of TableColumn objects
                or dictionaries representing table columns.
        """
        if self._columns and not refresh:
            return self._columns
        datasource_id = (
            self.datasource.id
            if isinstance(self.datasource, DatasourceInstance) else self.datasource.get("id")
        )
        columns = fetch_objects(
            connection=self.connection,
            api=datasources.get_table_columns,
            limit=None,
            filters={},
            dict_unpack_value="columns",
            datasource_id=datasource_id,
            namespace_id=self.namespace_id,
            table_id=self.__id,
        )

        if to_dictionary:
            return columns

        column_objects = TableColumn.bulk_from_dict(
            source_list=columns, connection=self.connection
        )
        if refresh:
            self._columns = column_objects
        return column_objects

    @property
    def dependent_logical_tables(self):
        if not self._dependent_logical_tables:
            self._dependent_logical_tables = self.list_dependent_logical_tables()
        return self._dependent_logical_tables

    @cached_property
    def columns(self):
        self._columns = self.list_columns()
        return self._columns



"""This is the demo script to show how to manage datasources.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.datasources import (
    DatasourceConnection,
    DatasourceInstance,
    DatasourceLogin,
    DatasourceMap,
    DatasourceType,
    ExecutionMode,
    list_available_dbms,
    list_datasource_connections,
    list_datasource_instances,
    list_datasource_logins,
    list_datasource_mappings
)
from mstrio.connection import get_connection

PROJECT_NAME = '<Project_name>'  # Insert project name here

# create a datasource login - all following information are concerning datasource logins management

LOGIN_NAME = '<Login_name>'  # Insert name for login to datasouce here
NEW_LOGIN_NAME = '<Login_name>'  # Insert new name for login to datasouce here
DATASOURCE_USER_NAME = '<DS_User_name>'  # insert name of user in DataSource here
DATASOURCE_PASSWORD = '<DS_Password>'  # Insert datasource password here
DATASOURCE_LOGIN_DESCRIPTION = '<DS_Login_Desc>'  # Insert datasource login description
DATASOURCE_LOGIN_NEW_DESCRIPTIOM = '<DS_Login_Desc>'  # Insert new datasource login description

# Following vaiables are for Datasource Connections management
DATASOURCE_CONNECTION_NAME = '<DS_Connection_Name>'  # insert name of datasource connection here
DATASOURCE_CONNECTION_DESCRIPTION = '<DS_Connection_Desc>'  # 'Insert description of datasource connection here'
DATASOURCE_CONNECTION_NEW_NAME = '<DS_Connection_Name>'  # insert new name of datasource connection if you wish to change it
DATASOURCE_CONNECTION_NEW_DESCRIPTION = '<DS_Connection_Desc>'  # insert new description of datasource connection if you wish to change it

# Following variables are for DataBase Management Systems (DMBSs)
DBMS_NAME = '<DBMS_Name>'  # Insert name of DBMS that you want to find

# Following variables are for Datasource Instances management
DATASOURCE_INSTANCE_NAME = '<DS_Instance_Name>'  # Insert name for created datasource instance here
DATASOURCE_INSTANCE_DESCRIPTION = '<DS_Instance_Desc>'  # Insert description for created datasource instance here
DATASOURCE_INSTANCE_TABLE_PREFIX = '<DS_Instance_Prefix>'  # Insert table prefix for created datasource instance here
DATASOURCE_INSTANCE_NEW_NAME = '<DS_Instance_Name>'  # Insert name for edited datasource instance here
DATASOURCE_INSTANCE_NEW_DESCRIPTION = '<DS_Instance_Desc>'  # Insert name for edited datasource instance here
DATASOURCE_INSTANCE_NEW_TABLE_PREFIX = '<DS_Instance_Prefix>'  # Insert new table prefix for edited datasource instance here

# Following variables are for Datasource Mappings
DATASOURCE_MAP_ID = '<DS_Map_ID>'  # Insert ID of datasource map that you wish to update
DATASOURCE_MAP_USERNAME = '<DS_Map_Username>'  # Insert user name that you wish to be mapped with this datasource mapping

conn = get_connection(workstationData, project_name=PROJECT_NAME)

login = DatasourceLogin.create(
    connection=conn,
    name=LOGIN_NAME,
    username=DATASOURCE_USER_NAME,
    password=DATASOURCE_PASSWORD,
    description=DATASOURCE_LOGIN_DESCRIPTION
)

# update a datasource login
login.alter(name=NEW_LOGIN_NAME, description=DATASOURCE_LOGIN_NEW_DESCRIPTIOM)

# list properties of a datasource login
login.list_properties()

# delete a datasource login
login.delete(force=True)

# list all datasource logins
list_datasource_logins(connection=conn)

# create a datasource connection
ds_conn = DatasourceConnection.create(
    connection=conn,
    name=DATASOURCE_CONNECTION_NAME,
    description=DATASOURCE_CONNECTION_DESCRIPTION,
    # the ExecutionMode values can be found in  datasources/datasource_connection.py
    execution_mode=ExecutionMode.SYNCHRONOUS,
    datasource_login=login
)

# update a datasource connection
ds_conn.alter(
    name=DATASOURCE_CONNECTION_NEW_NAME, description=DATASOURCE_CONNECTION_NEW_DESCRIPTION
)

# list properties of a datasource connection
ds_conn.list_properties()

# delete a datasource connection
ds_conn.delete(force=True)

# list all datasource connections
list_datasource_connections(connection=conn)

# get a dbms by name
dbms = list_available_dbms(connection=conn, name=DBMS_NAME)

# create a datasource instance
ds_instance = DatasourceInstance.create(
    connection=conn,
    name=DATASOURCE_INSTANCE_NAME,
    description=DATASOURCE_INSTANCE_DESCRIPTION,
    dbms=dbms,
    datasource_connection=ds_conn,
    table_prefix=DATASOURCE_INSTANCE_TABLE_PREFIX,
    # the DatasourceType values can be found in  datasources/datasource_instance.py
    datasource_type=DatasourceType.RESERVED
)

# update a datasource instance
ds_instance.alter(
    name=DATASOURCE_INSTANCE_NEW_NAME,
    description=DATASOURCE_INSTANCE_NEW_DESCRIPTION,
    table_prefix=DATASOURCE_INSTANCE_NEW_TABLE_PREFIX
)

# list properties of a datasource instance
ds_instance.list_properties()

# delete a datasource instance
ds_instance.delete(force=True)

# list all datasources by environment
list_datasource_instances(connection=conn)

# list all datasources by project
list_datasource_instances(connection=conn, project=PROJECT_NAME)

# list all datasources by datasource connection
list_datasource_instances(connection=conn, datasource_connection={"id": ds_conn.id})

# create a datasource map
ds_map = DatasourceMap.create(
    connection=conn,
    project=PROJECT_NAME,
    user=DATASOURCE_MAP_USERNAME,
    ds_connection=ds_conn,
    datasource=ds_instance,
    login=login
)

# initialise a datasource map
ds_map = DatasourceMap(connection=conn, id=DATASOURCE_MAP_ID)

# list all datasource maps
list_datasource_mappings(connection=conn)

# delete a datasource map
ds_map.delete(force=True)



# flake8: noqa
from .objects_and_references import (
    AttributeElement,
    AttributeFormNormalRef,
    AttributeFormSystemRef,
    AttributeRef,
    ElementPromptRef,
    FilterRef,
    ObjectInformation,
    ObjectReference
)
from .predicate_parameters import (
    ConstantArrayParameter,
    ConstantParameter,
    ConstantType,
    DynamicDateTimeParameter,
    ExpressionParameter,
    ObjectReferenceParameter,
    ParameterBase,
    ParameterType,
    PromptParameter
)
from .predicates import (
    LogicFunction,
    LogicOperator,
    PredicateBase,
    PredicateElementList,
    PredicateElementListFunction,
    PredicateFilter,
    PredicateForm,
    PredicateFormFunction,
    PredicateJointElementList
)
from .qualification import Qualification
from .security_filter import list_security_filters, SecurityFilter



"""This is the sample script to show how to manage schema facts.
This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to ease
its usage.
"""

from mstrio.connection import Connection, get_connection
from mstrio.modeling.expression import (
    ColumnReference, Expression, ExpressionFormat, FactExpression
)
from mstrio.modeling.schema.fact import Fact, list_facts
from mstrio.modeling.schema.helpers import DataType, ObjectSubType
from mstrio.modeling.schema import SchemaManagement, SchemaObjectReference, SchemaUpdateType
from workflows.get_all_columns_in_table import list_table_columns

# Following variables are defining basic facts
PROJECT_NAME = '<Project_name>'  # Insert name of project here
FACT_NAME = '<Fact_name>'  # Insert name of existing fact here
FACT_BY_NAME = '<Fact_unique_name>'  # Insert unique name of existing fact here
FACT_ID1 = '<Fact_ID>'  # Insert ID of existing fact here
FACT_ID2 = '<Fact_ID>'  # Insert ID of existing fact here
FACT_NEW_NAME = '<Fact_name>'  # Insert new name of edited Fact here
FACT_NEW_DESC1 = '<Fact_desc>'  # Insert new description of edited fact here
FACT_NEW_DESC2 = '<Fact_desc>'  # Insert new description of edited fact here
FOLDER_ID = '<Folder_ID>'  # Insert folder ID here

conn: Connection = get_connection(workstationData, PROJECT_NAME)

# Example fact data used to create new fact specified as dict
FACT_DATA = {
    'name': 'test_fact',
    'sub_type': ObjectSubType.FACT,
    'destination_folder': FOLDER_ID,
    'data_type': {
        'type': 'float', 'precision': 8, 'scale': -2147483648
    },
    'expressions': [
        {
            'expression': {
                'tokens': [{
                    'value': 'TOT_DOLLAR_SALES + TOT_COST'
                }]
            },
            'tables': [
                {
                    'objectId': '<Object_ID>', 'subType': 'logical_table', 'name': 'CITY_MNTH_SLS'
                }, {
                    'objectId': '<Object_ID>', 'subType': 'logical_table', 'name': 'CUSTOMER_SLS'
                }
            ]
        }
    ]
}

# fact expression data with expression specified as tree
FACT_EXP_DATA = FactExpression(
    expression=Expression(tree=ColumnReference(
        column_name='day_date',
        object_id='<Object_ID>',
    )),
    tables=[
        SchemaObjectReference(
            name='LU_DAY',
            sub_type=ObjectSubType.LOGICAL_TABLE,
            object_id='<Object_ID>',
        ),
    ],
)

# DataType object used to alter fact's data_type value
DATA_TYPE_DATA = DataType(type='float', precision=6, scale=-2147483648)

# Facts management
# Get list of facts, with examples of different conditions
list_of_all_facts = list_facts(connection=conn)
list_of_limited_facts = list_facts(connection=conn, limit=10)
list_of_limited_facts_to_dict = list_facts(connection=conn, limit=10, to_dictionary=True)
list_of_facts_by_name = list_facts(connection=conn, name=FACT_NAME)

# list of facts with expressions represented as trees
list_of_facts_as_trees = list_facts(connection=conn, show_expression_as=ExpressionFormat.TREE)

# list of facts with expressions respresented as tokens
list_of_facts_as_tokens = list_facts(connection=conn, show_expression_as=ExpressionFormat.TOKENS)

# create new fact
test_fact = Fact.create(
    connection=conn,
    name=FACT_DATA['name'],
    destination_folder=FACT_DATA['destination_folder'],
    expressions=FACT_DATA['expressions'],
    data_type=FACT_DATA['data_type']
)

# Any changes to a schema objects must be followed by schema_reload
# in order to use them in reports, dossiers and so on
schema_manager = SchemaManagement(connection=conn, project_id=conn.project_id)
task = schema_manager.reload(update_types=[SchemaUpdateType.LOGICAL_SIZE])

# Add new expression to the fact
test_fact.add_expression(FACT_EXP_DATA)
first_exp_id = test_fact.expressions[0].id

# Remove first expression from the fact
test_fact.remove_expression(first_exp_id)

# Alter fact
test_fact.alter(name=FACT_NEW_NAME, description=FACT_NEW_DESC1)
test_fact.alter(data_type=DATA_TYPE_DATA)

# Delete newly created fact
test_fact.delete(force=True)

# Get specific fact by id with expressions represented as trees
fact = Fact(connection=conn, id=FACT_ID1)

# List all tables for fact
tables_all = fact.get_tables()

# List tables for fact expression
exp_id = fact.expressions[0].id
exp_obj = fact.expressions[0]
tables_exp_id = fact.get_tables(expression=exp_id)
tables_exp_obj = fact.get_tables(expression=exp_obj)

# List fact properties
properties = fact.list_properties()

# Get existing fact by its unique name
fact_by_name = Fact(connection=conn, name=FACT_BY_NAME)

# Get specified fact by id with expressions represented as tokens
fact2 = Fact(connection=conn, id=FACT_ID2, show_expression_as=ExpressionFormat.TOKENS)

# Alter fact description
fact2.alter(description=FACT_NEW_DESC2)



from typing import Optional

from mstrio.utils.helper import Dictable


class Node(Dictable):
    _DELETE_NONE_VALUES_RECURSION = False

    def __init__(
        self, name: str, address: Optional[str] = None, service_control: Optional[bool] = None
    ) -> None:
        self.name = name
        self.address = address
        self.service_control = service_control



"""This is the demo script meant to show how administrator can perform a
migration of objects from one environment to another.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.

`mstrio.object_management.migration` module is available as a Functionality Preview.
It is subject to change until it is released as Generally Available.
"""

from mstrio.access_and_security.privilege import Privilege
from mstrio.connection import Connection
from mstrio.object_management.migration import (
    bulk_full_migration,
    bulk_migrate_package,
    Migration,
    PackageConfig,
    PackageContentInfo,
    PackageSettings
)
from mstrio.types import ObjectTypes
from mstrio.users_and_groups.user import User

# connection parameters
PROJECT_NAME = '<Project_name>'
SOURCE_BASE_URL = '<connection_base_url>'  # usually ends with /MicroStrategyLibrary/api
SOURCE_USERNAME = '<username>'
SOURCE_PASSWORD = '<password>'
TARGET_BASE_URL = '<connection_base_url>'  # usually ends with /MicroStrategyLibrary/api
TARGET_USERNAME = '<username>'
TARGET_PASSWORD = '<password>'

USER_NAME = '<username>'  # user that is executing the migration

# IDs for items to migrate
DOSSIER_ID = '<Dossier_id>'
REPORT_ID = '<Report_id>'

SAVE_PATH = 'path/to/import_package.mmp'
CUSTOM_PACKAGE_PATH = 'path/to/other_import_package.mmp'
UNDO_PACKAGE_PATH = 'path/to/undo_package.mmp'

# Create connections to both source and target environments
source_conn = Connection(
    SOURCE_BASE_URL, SOURCE_USERNAME, SOURCE_PASSWORD, project_name=PROJECT_NAME, login_mode=1
)
target_conn = Connection(
    TARGET_BASE_URL, TARGET_USERNAME, TARGET_PASSWORD, project_name=PROJECT_NAME, login_mode=1
)

# Make sure the current user have the following privileges:
#   'Create package', id: 295
#   'Apply package',  id: 296
# They can be granted by admin with the following commands:
user = User(source_conn, username=USER_NAME)
Privilege(source_conn, id=295).add_to_user(user)
Privilege(source_conn, id=296).add_to_user(user)

# Or by name:
user2 = User(target_conn, username=USER_NAME)
Privilege(target_conn, name='Create package').add_to_user(user2)
Privilege(target_conn, name='Apply package').add_to_user(user2)

# Create PackageConfig with information what object should be migrated and how.
# The options are of type Enum with all possible values listed.
package_settings = PackageSettings(
    PackageSettings.DefaultAction.USE_EXISTING,
    PackageSettings.UpdateSchema.RECAL_TABLE_LOGICAL_SIZE,
    PackageSettings.AclOnReplacingObjects.REPLACE,
    PackageSettings.AclOnNewObjects.KEEP_ACL_AS_SOURCE_OBJECT,
)
package_content_info = PackageContentInfo(
    id=REPORT_ID,
    type=ObjectTypes.REPORT_DEFINITION,
    action=PackageContentInfo.Action.USE_EXISTING,
    include_dependents=True,
)
package_content_info2 = PackageContentInfo(
    id=DOSSIER_ID,
    type=ObjectTypes.DOCUMENT_DEFINITION,
    action=PackageContentInfo.Action.USE_EXISTING,
    include_dependents=True,
)

package_config = PackageConfig(
    PackageConfig.PackageUpdateType.PROJECT, package_settings, package_content_info
)
package_config2 = PackageConfig(
    PackageConfig.PackageUpdateType.PROJECT, package_settings, package_content_info2
)

# Create Migrations objects that can use all the functionalities
mig = Migration(
    save_path=SAVE_PATH,
    source_connection=source_conn,
    target_connection=target_conn,
    configuration=package_config,
)

# Create Migration object that can only use `create_package()`
mig2 = Migration(
    save_path=SAVE_PATH,
    source_connection=source_conn,
    configuration=package_config2,
)

# Create Migration object that can only be used for `migrate_package()`
mig3 = Migration(
    target_connection=target_conn,
    custom_package_path=CUSTOM_PACKAGE_PATH,
)

# Short version
# Create import package and save it to the file
mig.create_package()
# or
mig2.create_package()

# Migrate downloaded package to the target environment.
# Create undo package and save it to file
mig.migrate_package()
# or
mig3.migrate_package()

# End to end migration
mig.perform_full_migration()

# Detailed version
# Create import package and save it to the file specified with `save_path`
# argument during creation of migration object
mig.create_package()
mig2.create_package()

# Migrate downloaded package to the target environment
# `migrate_package()` by default uses a package binary saved to a variable
# during `create_package()`
mig.migrate_package()

# or a custom package binary specified with `custom_package_path`
# during Migration object creation, if `create_package()` was not called.
mig3.migrate_package()

# It is also possible to respecify `custom_package_path` at this stage
mig3.migrate_package(custom_package_path=CUSTOM_PACKAGE_PATH)

# Perform end to end migration. `perform_full_migration()` encapsulates
# `create_package()` and `migrate_package()` from the previous steps
# In order to be able to use, Migration object needs `source_connection`,
# `configuration` and `target_connection` parameters filled during creation
mig.perform_full_migration()

# Perform many full migrations at once
bulk_full_migration([mig])

# Perform many migrations at once
bulk_migrate_package([mig, mig3])

# If the migration needs to be reverted use `undo_migration()`
mig.undo_migration()

# or run `migrate_package()` with path to the custom undo package
Migration(
    save_path=SAVE_PATH, target_connection=target_conn, custom_package_path=UNDO_PACKAGE_PATH
).migrate_package()

# Status of the migration can be checked by checking the `status` property
status = mig.status



# flake8: noqa
from .cluster import Cluster, GroupBy, ServiceAction
from .environment import Environment
from .node import Node
from .project import compare_project_settings, IdleMode, Project, ProjectSettings, ProjectStatus
from .server import ServerSettings

# isort: off
from .job_monitor import (
    Job,
    JobStatus,
    JobType,
    kill_all_jobs,
    kill_jobs,
    list_jobs,
    ObjectType,
    PUName,
    SubscriptionType
)
# isort: on



"""This is the demo script to show how administrator can manage users and
user groups.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""
import csv

from mstrio.users_and_groups import (
    create_users_from_csv, list_user_groups, list_users, User, UserGroup
)

from mstrio.connection import get_connection

PROJECT_NAME = '<project_name>'  # Project to connect to

# Usernames and full names for users to create
USERNAME_1 = '<username_1>'
FULLNAME_1 = '<full_name_1>'
USERNAME_2 = '<username_2>'
FULLNAME_2 = '<full_name_2>'
USERNAME_3 = '<username_3>'
FULLNAME_3 = '<full_name_3>'
USERNAME_4 = '<username_4>'
FULLNAME_4 = '<full_name_4>'
USERNAME_5 = '<username_5>'
FULLNAME_5 = '<full_name_5>'
USERNAME_6 = '<username_6>'
FULLNAME_6 = '<full_name_6>'

NAME_BEGINS = '<name_begins>'  # beginning of the full name to look for in listing users
INITIALS = '<name_initials>'  # initials to look for in listing users
USERGROUP_NAME = '<usergroup_name>'  # name of a User Group to create and add users
CSV_FILE = '<path_to_csv_file>'  # csv file to create user from
OBJECT_ID = '<object_id>'  # ID of an object to which permissions will be set
OBJECT_TYPE = '<object_type>'  # type of an object to which permissions will be set;
# see ObjectTypes enum in types.py for available object type values
PERMISSION = '<permission_type>'  # type of permission to set;
# see Permissions enum in /utils/acl.py for available permission values

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# create multiple users
users_array = [
    {
        'username': USERNAME_1, 'fullName': FULLNAME_1
    }, {
        'username': USERNAME_2, 'fullName': FULLNAME_2
    }, {
        'username': USERNAME_3, 'fullName': FULLNAME_3
    }, {
        'username': USERNAME_4, 'fullName': FULLNAME_4
    }, {
        'username': USERNAME_5, 'fullName': FULLNAME_5
    }
]
for u in users_array:
    User.create(connection=conn, username=u['username'], full_name=u['fullName'])

# Also, you can create users from a CSV file
newly_created_users = create_users_from_csv(connection=conn, csv_file=CSV_FILE)

# Or you can do it manually
with open(CSV_FILE, "r") as f:
    users = csv.DictReader(f)

    for user in users:
        User.create(connection=conn, username=user['username'], full_name=user['full_name'])

# create a single user and get users which name begins with "John" and have
# additional filter for initials
User.create(connection=conn, username=USERNAME_6, full_name=FULLNAME_6)
my_users = list_users(connection=conn, name_begins=NAME_BEGINS, initials=INITIALS)

# get all user groups (you can also add additional filters as for users) and
# create a new one
user_groups_list = list_user_groups(connection=conn)
UserGroup.create(connection=conn, name=USERGROUP_NAME)

# get user, user group and add this user to this user group
user_ = User(connection=conn, name=FULLNAME_6)
user_group_ = UserGroup(connection=conn, name=USERGROUP_NAME)
user_group_.add_users(users=[user_.id])

# set custom permissions of the user for given objects
user_.set_custom_permissions(
    to_objects=[OBJECT_ID],
    object_type=OBJECT_TYPE,
    execute=
    '<permission>',  # Available values for permissions are 'grant', 'deny', 'default' or None
    use='<permission>',
    control='<permission>',
    delete='<permission>',
    write='<permission>',
    read='<permission>',
    browse='<permission>'
)

# set permission of the user group for given objects
user_group_.set_permission(permission=PERMISSION, to_objects=[OBJECT_ID], object_type=OBJECT_TYPE)

# delete a user and a user group
user_.delete()
user_group_.delete()

# Addresses
user_john = User(connection=conn, name=FULLNAME_6)
# The user's addresses list is lazily loaded upon first accessing the property
johns_addresses = user_john.addresses
# MicroStrategy allows having multiple addresses marked as default
# so let's construct a list of John's "default" addresses
johns_default_addresses = [addr for addr in johns_addresses if addr["is_default"]]
# If we just want a single address to treat as default and we're OK with
# it being just the first "default" one we can now just take the first element
# of the list
johns_default_address = johns_default_addresses[0]
# Alternatively if we don't care about any addresses besides the first
# "default one" we can use a generator expression and take its first output
# to avoid going through the entire list, increasing performance
johns_default_address = next((addr for addr in user_john.addresses if addr["is_default"]), None)



from mstrio.connection import Connection
from mstrio.utils.api_helpers import changeset_decorator
from mstrio.utils.error_handlers import ErrorHandler


@changeset_decorator
@ErrorHandler(err_msg='Error getting attribute {id} relationship.')
def get_attribute_relationships(
    connection: Connection, id: str, changeset_id: str, project_id: str
):
    """Get relationship(s) of an attribute

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of a attribute
        changeset_id: ID of a changeset
        project_id: Id of a project

    Return:
        HTTP response object. Expected status: 200
    """
    return connection.get(
        url=f'{connection.base_url}/api/model/systemHierarchy/attributes/{id}/relationships',
        headers={
            'X-MSTR-MS-Changeset': changeset_id, 'X-MSTR-ProjectID': project_id
        }
    )


@changeset_decorator
@ErrorHandler(err_msg='Error updating attribute {id} relationship.')
def update_attribute_relationships(
    connection: Connection,
    id: str,
    changeset_id: str,
    body: dict,
):
    """Update relationship(s) of an attribute

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of a attribute
        changeset_id: ID of a changeset
        body: JSON-formatted definition of the attribute relationships

    Return:
        HTTP response object. Expected status: 200
    """
    return connection.put(
        url=f'{connection.base_url}/api/model/systemHierarchy/attributes/{id}/relationships',
        headers={'X-MSTR-MS-Changeset': changeset_id},
        json=body
    )



"""This is the demo script to show how to manage contact groups.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

# create contact group with one member represented as an object (it can be more
# and they can be also represented as dictionaries). User linked to the contact
# group can be passed as a `User` object or just its ID.

from mstrio.distribution_services import (ContactGroup, ContactGroupMember, ContactGroupMemberType)
from mstrio.users_and_groups import User
from mstrio.connection import get_connection
from production.mstrio.distribution_services.contact.contact import Contact
from production.mstrio.distribution_services.contact_group.contact_group import list_contact_groups

PROJECT_NAME = '<Project_name>'  # Insert project name here
MEMBER_NAME = '<Username>'  # here insert member name
MEMBER_ID = '<Member_ID>'  # Insert member ID
USER_NAME = '<Username>'  # name of user to link to the contact group
CONTACT_GROUP_NAME = '<Contact_group_name>'  # Name of contact group
CONTACT_GROUP_DESCRIPTION = '<Contact_group_desc>'  # Description of newly created contact group
CONTACT_GROUP_ID = '<Contact_group_ID>'  # Insert your contact group ID here
CONTACT_GROUP_NEW_NAME = '<Contact_group_name>'  # Insert new name of contact group
CONTACT_GROUP_NEW_DESCRIPTION = '<Contact_group_desc>'  # Insert new description of contact group
OTHER_USER_NAME = '<Username>'  # Insert new user for contact group
CONTACT_NAME = '<Contact_name>'  # Insert contact name here
OTHER_CONTACT_GROUP_NAME = '<Contact_group_name>'  # Insert name of another contact group here

conn = get_connection(workstationData, project_name=PROJECT_NAME)

member1 = ContactGroupMember(
    name=MEMBER_NAME,
    type=ContactGroupMemberType.CONTACT,  # it can be also ContactGroupMemberType.CONTACT_GROUP
    id=MEMBER_ID,
    enabled=True,
)
new_cg = ContactGroup.create(
    connection=conn,
    name=CONTACT_GROUP_NAME,
    linked_user=User(conn, username=USER_NAME),
    members=[member1],
    description=CONTACT_GROUP_DESCRIPTION,
    enabled=True,
)

# list all contact groups
contact_groups = list_contact_groups(conn)

# get contact group by ID. Contact group can be also found by its name
cg = ContactGroup(conn, id=CONTACT_GROUP_ID)
cg_by_name = ContactGroup(conn, name=CONTACT_GROUP_NAME)

# get properties of contact group
cg_name = cg.name
cg_description = cg.description
cg_lined_user = cg.linked_user
cg_members = cg.members
cg_enabled = cg.enabled

# alter contact group
cg.alter(
    name=CONTACT_GROUP_NEW_NAME,
    description=CONTACT_GROUP_NEW_DESCRIPTION,
    linked_user=User(conn, username=OTHER_USER_NAME),
    enabled=False,
)

# add members to contact group - it can be `ContactGroupMember`, `Contact`
# or `ContactGroup`
contact = Contact(conn, name=CONTACT_NAME)
cg.add_members(members=[contact])  # add contact
cg.add_members(members=[new_cg])  # add contact group

# remove members from contact group - it can be `ContactGroupMember`, `Contact`
# or `ContactGroup`
contact_group = ContactGroup(conn, name=OTHER_CONTACT_GROUP_NAME)
cg.remove_members(members=[contact])  # remove contact
cg.remove_members(members=[contact_group])  # remove contact group

# Delete contact group. When argument `force` is set to `False` (default value),
# then deletion must be confirmed by selecting appropriate prompt value.
cg.delete(force=True)



# flake8: noqa
from .contact import *
from .contact_group import *
from .device import *
from .event import Event, list_events
from .schedule import *
from .subscription import *
from .transmitter import *



from typing import Optional

from mstrio.connection import Connection
from mstrio.utils.api_helpers import changeset_decorator, unpack_information
from mstrio.utils.error_handlers import ErrorHandler


@unpack_information
@ErrorHandler(err_msg='Error reading fact with ID: {id}.')
def read_fact(
    connection: "Connection",
    id: str,
    project_id: str = None,
    changeset_id: str = None,
    show_expression_as: Optional[str] = None,
    show_potential_tables: bool = False,
    show_fields: Optional[str] = None
):
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id
    spt = str(show_potential_tables).lower() if show_potential_tables is not None else None
    return connection.get(
        url=f"{connection.base_url}/api/model/facts/{id}",
        headers={
            'X-MSTR-ProjectID': project_id,
            'X-MSTR-MS-Changeset': changeset_id,
        },
        params={
            'showExpressionAs': show_expression_as,
            'showPotentialTables': spt,
            'showFields': show_fields,
        },
    )


@unpack_information
@changeset_decorator
@ErrorHandler(err_msg='Error creating a fact.')
def create_fact(
    connection: "Connection",
    changeset_id: str,
    body: dict,
    show_expression_as: Optional[str] = None,
    show_potential_tables: bool = False
):
    spt = str(show_potential_tables).lower() if show_potential_tables is not None else None
    return connection.post(
        url=f"{connection.base_url}/api/model/facts",
        headers={'X-MSTR-MS-Changeset': changeset_id},
        params={
            'showExpressionAs': show_expression_as,
            'showPotentialTables': spt,
        },
        json=body,
    )


@unpack_information
@changeset_decorator
@ErrorHandler(err_msg='Error updating fact with ID: {id}.')
def update_fact(
    connection: "Connection",
    changeset_id: str,
    id: str,
    body: dict,
    show_expression_as: Optional[str] = None,
    show_potential_tables: bool = False
):
    spt = str(show_potential_tables).lower() if show_potential_tables is not None else None
    return connection.put(
        url=f"{connection.base_url}/api/model/facts/{id}",
        headers={'X-MSTR-MS-Changeset': changeset_id},
        params={
            'showExpressionAs': show_expression_as,
            'showPotentialTables': spt,
        },
        json=body,
    )



from dataclasses import dataclass
from enum import auto
import logging
from typing import Optional, TYPE_CHECKING

from mstrio.modeling.schema.helpers import SchemaObjectReference
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import Dictable
from mstrio.utils.wip import module_wip, WipLevels

module_wip(globals(), level=WipLevels.WARNING)
logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from mstrio.connection import Connection


@dataclass
class DimensionalityUnit(Dictable):
    """
    Object that specifies the dimensionality unit.

    Attributes:
        dimtyUnitType: Type of unit in dimensionality
        target: target of dimensionality
        aggregation: type of aggregation in dimensionality
        filtering: Type of filtering applied in this dimensionality
        relativePosition: number showing relative position of dimensionality
        in metric reports
        target: Object reference to attribute or report, on which the
        dimensionality will be grouped at
        axisCollection: dict of numbers showing, at which axis
        metric will be shown
    """

    class DimensionalityUnitType(AutoName):
        DEFAULT = auto()
        ATTRIBUTE = auto()
        DIMENSION = auto()
        REPORT_LEVEL = auto()
        REPORT_BASE_LEVEL = auto()
        ROLE = auto()

    class Aggregation(AutoName):
        NORMAL = auto()
        FIRST_IN_FACT = auto()
        LAST_IN_FACT = auto()
        FIRST_IN_RELATIONSHIP = auto()
        LAST_IN_RELATIONSHIP = auto()

    class Filtering(AutoName):
        APPLY = auto()
        ABSOLUTE = auto()
        IGNORE = auto()
        NONE = auto()

    _DELETE_NONE_VALUES_RECURSION = False
    _FROM_DICT_MAP = {
        'dimensionality_unit_type': DimensionalityUnitType,
        'filtering': Filtering,
        'aggregation': Aggregation,
        'target': SchemaObjectReference.from_dict
    }

    dimensionality_unit_type: DimensionalityUnitType
    aggregation: Optional[Aggregation] = None
    filtering: Optional[Filtering] = None
    group_by: Optional[bool] = None
    relative_position: Optional[int] = None
    target: Optional[SchemaObjectReference] = None
    axis_collection: Optional[dict] = None

    @classmethod
    def from_dict(cls, source: dict):
        new_source = source.copy()
        new_source['dimensionality_unit_type'] = source.get('dimtyUnitType', [])
        return super().from_dict(new_source)

    def to_dict(cls, camel_case: bool = True) -> dict:
        obj_dict = super().to_dict(camel_case=camel_case)
        if obj_dict.get('dimensionalityUnitType'):
            obj_dict['dimtyUnitType'] = obj_dict.pop('dimensionalityUnitType')
        return obj_dict


@dataclass
class Dimensionality(Dictable):
    """
    Object that specifies the dimensionality.

    Attributes:
        dimtyUnits: The list of dimty unit
        excludeAttribute: boolean whether to exclude attribute
            absent in report or level(dimensionality)
        allowAddingUnit: boolean whether to allow other users
            to add extra units
        prompt: reference to a prompt
    """

    _DELETE_NONE_VALUES_RECURSION = False

    _FROM_DICT_MAP = {
        'dimensionality_units': [DimensionalityUnit], 'prompt': SchemaObjectReference
    }

    dimensionality_units: Optional[list[DimensionalityUnit]] = None
    exclude_attribute: Optional[bool] = None
    allow_adding_unit: Optional[bool] = None
    prompt: Optional[SchemaObjectReference] = None

    @classmethod
    def from_dict(
        cls, source: dict, connection: Optional['Connection'] = None, to_snake_case: bool = True
    ):
        data = source.copy()
        data['dimensionality_units'] = data.get('dimty_units', [])
        return super().from_dict(data, connection, to_snake_case)

    def to_dict(cls, camel_case: bool = True) -> dict:
        obj_dict = super().to_dict(camel_case=camel_case)
        if obj_dict.get('dimensionalityUnits'):
            obj_dict['dimtyUnits'] = obj_dict.pop('dimensionalityUnits')
        return obj_dict



"""This is the demo script to show how to manage devices.
This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.distribution_services import (
    BackupPrinterProperties,
    ConnectionParameters,
    Device,
    DeviceType,
    EmailDeviceProperties,
    EmailFormat,
    EmailSmartHostSettings,
    FileDeviceProperties,
    FileLocation,
    FileProperties,
    FileSystem,
    IOSDeviceProperties,
    list_devices,
    PrinterDeviceProperties,
    PrinterLocation,
    PrinterProperties,
    PrinterPdfSettings,
    Transmitter,
    UnixWindowsSharity
)
from mstrio.connection import get_connection

PROJECT_NAME = '<Project_name>'  # Project to connect to
SMART_HOST_SERVER = '<example.com>'  # server name for smart host
SMART_HOST_PORT = 22  # server port for smart host
FEEDBACK_SERVICE_SERVER = '<Service_name>'  # server name for ios feedback server
FEEDBACK_SERVICE_PORT = 1212  # server port for ios feedback server
FILE_DEVICE_PATH = 'path/to/file'  # path for file device
PRINTER_LOCATION = '<Printer_location>'  # printer location field for printer device properties
TRANSMITTER_ID = '<Transmitter_id>'  # id for Transmitter object
DEVICE_ID = '<Device_id>'  # id for Device object
DEVICE_NAME = '<Name_of_device>'
DEVICE_DESCRIPTION = '<Description_of_device>'

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# create email device properties object
# which is used when creating device
edp = EmailDeviceProperties(
    format=EmailFormat.UU_ENCODED,
    smart_host_settings=EmailSmartHostSettings(server=SMART_HOST_SERVER, port=SMART_HOST_PORT),
)
# see distribution_services/device/device_properties.py for EmailFormat values

# create ios device properties object
# which is used when creating iphone or ipad device
idp = IOSDeviceProperties(
    app_id='com.microstrategy.dossier.iphone',
    server='api.push.apple.com',
    port=443,
    feedback_service_server=FEEDBACK_SERVICE_SERVER,
    feedback_service_port=FEEDBACK_SERVICE_PORT,
)

# create file device properties object
# which is used when creating file device
file_location = FileLocation(file_path=FILE_DEVICE_PATH)
file_system = FileSystem(create_folder=False)
fp = FileDeviceProperties(
    file_location=file_location,
    file_system=file_system,
    connection_parameters=ConnectionParameters(retries_count=6),
    file_properties=FileProperties(read_only=False),
    unix_windows_sharity=UnixWindowsSharity(sharity_enabled=False),
)

# create print device properties object
# which is used when creating print device
printer_properties = PrinterProperties(pdf_setting=PrinterPdfSettings(post_script_level=1))
pdp = PrinterDeviceProperties(
    printer_location=PrinterLocation(location=PRINTER_LOCATION),
    printer_properties=printer_properties,
    connection_parameters=ConnectionParameters(retries_count=6),
    backup_printer_properties=BackupPrinterProperties(print_on_backup=True),
)
# get transmitters by id
transmitter = Transmitter(conn, id=TRANSMITTER_ID)

# create a device with device type as `email` (when type is `email` then
# it is mandatory to provide `email_device_properties`)
new_email_device = Device.create(
    connection=conn,
    name=DEVICE_NAME,
    # the DeviceType values can be found in distribution_services/device/device.py
    device_type=DeviceType.EMAIL,
    description=DEVICE_DESCRIPTION,
    transmitter=transmitter,
    device_properties=edp,
)
# create a device with device type as `iphone` (when type is `iphone` then
# it is mandatory to provide `device_properties`). Device type
# can be also passed as a string and `device_properties` can be
# passed as a dictionary
new_iphone_device = Device.create(
    connection=conn,
    name=DEVICE_NAME,
    # the DeviceType values can be found in distribution_services/device/device.py
    device_type=DeviceType.IPHONE,
    description=DEVICE_DESCRIPTION,
    transmitter=transmitter,
    device_properties=idp,
)

# create a device with device type as `ipad`
new_ipad_device = Device.create(
    connection=conn,
    name=DEVICE_NAME,
    # the DeviceType values can be found in distribution_services/device/device.py
    device_type=DeviceType.IPAD,
    description=DEVICE_DESCRIPTION,
    transmitter=transmitter,
    device_properties=idp,
)

# create a device with device type as `file`
new_file_device = Device.create(
    connection=conn,
    name=DEVICE_NAME,
    # the DeviceType values can be found in distribution_services/device/device.py
    device_type=DeviceType.FILE,
    description=DEVICE_DESCRIPTION,
    transmitter=transmitter,
    device_properties=fp,
)

# create a device with device type as `printer`
new_printer_device = Device.create(
    connection=conn,
    name=DEVICE_NAME,
    # the DeviceType values can be found in distribution_services/device/device.py
    device_type=DeviceType.PRINTER,
    description=DEVICE_DESCRIPTION,
    transmitter=transmitter,
    device_properties=pdp,
)

# get list of devices
print(list_devices(conn))

# get device by ID. Device can be also found by its name.
device = Device(conn, id=DEVICE_ID)
device_by_name = Device(conn, name=DEVICE_NAME)

# alter email type device and its properties
edp.smart_host_settings.server = 'microstrategy.com'
edp.smart_host_settings.port = 21
new_email_device.alter(
    name=DEVICE_NAME,
    description=DEVICE_DESCRIPTION,
    device_properties=edp,
)

# list device properties
device.list_properties()

# Delete device. When argument `force` is set to `False` (default value),
# then deletion must be confirmed by selecting appropriate prompt value.
new_email_device.delete(force=True)
new_iphone_device.delete(force=True)



# flake8: noqa
# isort: off
from .package import Package, PackageConfig, PackageContentInfo, PackageImport, PackageSettings
from .migration import bulk_full_migration, bulk_migrate_package, Migration, MigrationStatus
# isort: on



"""This is the demo script to show how administrator can manage subscriptions
and schedules.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.distribution_services import (
    CacheType,
    CacheUpdateSubscription,
    Content,
    EmailSubscription,
    list_schedules,
    list_subscriptions,
    Schedule,
    Subscription,
    SubscriptionManager
)

from mstrio.connection import get_connection

PROJECT_NAME = '<project_name>'  # Project to connect to
CONTENT_TYPE = '<content_type>'  # see distribution_services/subscription/content.py for available options
FORMAT_TYPE = '<format_type>'  # see distribution_services/subscription/content.py for available options
EMAIL_SUBSCRIPTION_NAME = '<email_subscription_name>'
EMAIL_SUBJECT = '<email_subject>'
CACHE_SUBSCRIPTION_NAME = '<cache_subscription_name>'
SCHEDULE_NAME = '<schedule_name>'
DELIVERY_EXPIRATION_DATE = '<delivery_expiration_date>'

SUBSCRIPTION_ID = '<subscription_id>'
SUBSCRIPTION_ID_2 = '<subscription_id_2>'
CONTENT_ID = '<content_id>'
PROJECT_ID = '<project_id>'
RECIPIENT_ID = '<recipient_id>'
RECIPIENT_ID_2 = '<recipient_id_2>'
REMOVED_USER_ID = '<user_to_remove_id>'
ADMIN_USER_ID = '<admin_user_id>'
SCHEDULE_ID = '<schedule_id>'
OWNER_ID = '<owner_id>'

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Initialize manager for subscriptions on a chosen project
sub_mngr = SubscriptionManager(connection=conn, project_name=PROJECT_NAME)
# get all subscriptions from the given project (it is possible in two ways)
all_subs = list_subscriptions(connection=conn, project_name=PROJECT_NAME)
all_subs = sub_mngr.list_subscriptions()

#  execute/delete subscriptions by passing theirs ids or Subscription objects
sub_mngr.execute([SUBSCRIPTION_ID, SUBSCRIPTION_ID_2])
sub_mngr.delete([SUBSCRIPTION_ID, SUBSCRIPTION_ID_2], force=True)

# list available recipients of the subscription for the given content (default
# delivery type is an email)
sub_mngr.available_recipients(content_id=CONTENT_ID, content_type=CONTENT_TYPE)

# get a single subscription
sub = Subscription(connection=conn, subscription_id=SUBSCRIPTION_ID, project_id=PROJECT_ID)
# list all recipients of the given subscription and all available for this
# subscription
sub.recipients
sub.available_recipients()

# add/remove recipient(s) with given id(s)
sub.add_recipient(recipients=[RECIPIENT_ID, RECIPIENT_ID_2])
sub.remove_recipient(recipients=[RECIPIENT_ID, RECIPIENT_ID_2])

# execute a given subscription
sub.execute()

# replace a user with an admin in all of its subscriptions (e.g. when user exits
# company)
for s in sub_mngr.list_subscriptions(to_dictionary=False):
    if REMOVED_USER_ID in [r['id'] for r in s.recipients]:
        s.add_recipient(recipients=ADMIN_USER_ID)
        s.remove_recipient(recipients=REMOVED_USER_ID)

# create an email subscription
EmailSubscription.create(
    connection=conn,
    name=EMAIL_SUBSCRIPTION_NAME,
    project_name=PROJECT_NAME,
    contents=Content(
        id=CONTENT_ID,
        type=Content.Type
        .REPORT,  # see distribution_services/subscription/content.py for available options
        personalization=Content.Properties(format_type=FORMAT_TYPE)
    ),
    schedules=[SCHEDULE_ID],
    recipients=[RECIPIENT_ID],
    email_subject=EMAIL_SUBJECT
)

# create a cache update subscription
cache_update_sub = CacheUpdateSubscription.create(
    connection=conn,
    project_name=PROJECT_NAME,
    name=CACHE_SUBSCRIPTION_NAME,
    contents=Content(
        id=CONTENT_ID,
        type=Content.Type
        .REPORT,  # see distribution_services/subscription/content.py for available options
        personalization=Content.Properties(format_type=Content.Properties.FormatType.EXCEL),
    ),
    schedules=[SCHEDULE_ID],
    delivery_expiration_date=DELIVERY_EXPIRATION_DATE,
    send_now=True,
    recipients=[RECIPIENT_ID],
    cache_cache_type=CacheType.RESERVED
)  # see distribution_services/subscription/delivery.py for available options

# change name and owner of cache update subscription
cache_update_sub.alter(name=f"<{cache_update_sub.name}_(Altered)>", owner_id=OWNER_ID)

# list all cache update subscriptions
cache_update_subs = [
    sub for sub in list_subscriptions(conn, project_name=PROJECT_NAME)
    if isinstance(sub, CacheUpdateSubscription)
]

# get list of schedules (you can filter them by for example name, id or
# description)
all_schedules = list_schedules(conn)

# get a single schedule by its id or name and then its properties
schedule = Schedule(connection=conn, name=SCHEDULE_NAME)
schedule.list_properties()



"""This is the demo script to show how administrator can manage folders, objects
and their dependencies.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.object_management import (
    Folder,
    full_search,
    get_my_personal_objects_contents,
    get_predefined_folder_contents,
    get_search_results,
    get_search_suggestions,
    list_folders,
    list_objects,
    Object,
    PredefinedFolders,
    quick_search,
    quick_search_from_object,
    SearchObject,
    SearchPattern,
    SearchResultsFormat,
    start_full_search
)
from mstrio.project_objects import Dossier
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.utils.entity import Rights

from mstrio.connection import get_connection

PROJECT_NAME = '<Project_name>'  # Project to connect to
PROJECT_ID = '<Project_ID>'

# properties of folder to be created or altered
FOLDER_NAME = '<Name_of_folder>'
FOLDER_PARENT_ID = '<Folder_ID>'
FOLDER_DESCRIPTION = '<Description_of_folder>'
FOLDER_ID = '<Folder_ID>'  # ID for Folder object lookup
TRUSTED_USER_ID = '<User_ID>'

OBJECT_ID = '<Object_ID>'  # ID for object lookup
OBJECT_NAME = '<Name_of_object>'
OBJECT_DESCRIPTION = '<Description_of_object>'

SEARCH_PATTERN = '<pattern>'  # string to get search suggestions for

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# list folders from a particular project
list_folders(conn, project_id=PROJECT_ID)

# list configuration-level folders
list_folders(conn)

# get contents of My Personal Objects in a specific folder
get_my_personal_objects_contents(conn, project_id=PROJECT_ID)

# get contents of a pre-defined folder in a specific folder
get_predefined_folder_contents(
    conn, folder_type=PredefinedFolders.PUBLIC_REPORTS, project_id=PROJECT_ID
)

# create new folder in a particular folder
new_folder = Folder.create(
    conn, name=FOLDER_NAME, parent=FOLDER_PARENT_ID, description=FOLDER_DESCRIPTION
)

# get folder with a given ID
folder = Folder(conn, id=FOLDER_ID)

# get contents of a folder (optionally as a dictionary)
contents_objs = folder.get_contents()
contents_dict = folder.get_contents(to_dictionary=True)

# alter name and description of a folder
folder.alter(name=FOLDER_NAME, description=FOLDER_DESCRIPTION)

# add ACL right to the folder for the user and propagate to children of folder
folder.acl_add(
    Rights.EXECUTE,
    trustees=TRUSTED_USER_ID,
    denied=False,
    inheritable=True,
    propagate_to_children=True
)  # see utils/acl.py for Rights values

# remove ACL from the folder for the user
folder.acl_remove(Rights.EXECUTE, trustees=TRUSTED_USER_ID, denied=False, inheritable=True)
# see utils/acl.py for Rights values

# Delete folder. When argument `force` is set to `False` (default value), then
# deletion must be approved.
folder.delete(force=True)

# list objects of a given type
reports = list_objects(conn, ObjectTypes.REPORT_DEFINITION)  # see types.py for ObjectTypes values
documents = list_objects(conn, ObjectTypes.DOCUMENT_DEFINITION)

# initialize an object of a given type (in this case `FOLDER`) with a given id
object = Object(conn, ObjectTypes.FOLDER, OBJECT_ID)  # see types.py for ObjectTypes values

# alter name and description of an object
object.alter(name=OBJECT_NAME, description=OBJECT_DESCRIPTION)

# certify object
Object(conn, ObjectTypes.REPORT_DEFINITION, OBJECT_ID).certify()

# decertify object
Object(conn, ObjectTypes.REPORT_DEFINITION, OBJECT_ID).decertify()

# get object properties as a dictionary
object.to_dict()

# create copy of the object in the same folder with the default name
new_object = object.create_copy()

# create copy of the object in the same folder with provided name
new_object = object.create_copy(OBJECT_NAME)

# create copy of the object in another folder with provided name
new_object = object.create_copy(OBJECT_NAME, FOLDER_ID)

# move the object to another folder
object.move(FOLDER_ID)

# Delete objects of a given types (in this case `REPORT` and 'DOCUMENT)
# and given ids. When argument `force` is set to `False` (default value), then
# deletion must be approved.
Object(conn, ObjectTypes.REPORT_DEFINITION, OBJECT_ID).delete()
Object(conn, ObjectTypes.DOCUMENT_DEFINITION, OBJECT_ID).delete(force=True)

# initialize SearchObject and synchronize with server
search_object = SearchObject(conn, id=OBJECT_ID)

# get search suggestions with the pattern set performed in all unique projects
# across the cluster (it takes effect only in I-Server with cluster nodes)
suggestions = get_search_suggestions(
    conn, project=PROJECT_ID, key=SEARCH_PATTERN, max_results=20, cross_cluster=True
)

# use the stored results of the Quick Search engine to return search results
# and display them as a list (in this particular case all reports which name
# begins with 'A')
objects = quick_search(
    conn,
    PROJECT_ID,
    name='A',
    pattern=SearchPattern.BEGIN_WITH,
    object_types=[ObjectTypes.REPORT_DEFINITION]
)

# perform quick search based on a predefined Search Object (include ancestors
# and acl of returned objects)
objects = quick_search_from_object(
    conn, PROJECT_ID, search_object, include_ancestors=True, include_acl=True
)

# search the metadata for objects in a specific project that match specific
# search criteria (e.g. super cubes which name ends with `cube`) and save the
# results in I-Server memory
start_dict = start_full_search(
    conn,
    PROJECT_ID,
    name='cube',
    pattern=SearchPattern.END_WITH,
    object_types=ObjectSubTypes.SUPER_CUBE
)

# Retrieve the results of a full metadata search in a tree format which were
# previously obtained by `start_full_search`
results = get_search_results(
    conn, search_id=start_dict['id'], results_format=SearchResultsFormat.TREE
)

# perform full search in one call (instead of the two steps presented above).
# Get all documents which name contains `document` and return in a list format.
# Perform search only in the root folder and its subfolders.
results = full_search(
    conn,
    PROJECT_ID,
    name='document',
    pattern=SearchPattern.CONTAINS,
    object_types=ObjectTypes.DOCUMENT_DEFINITION,
    root=FOLDER_ID
)

# return cubes that are used by the given dossier (it can be performed with the
# function `full_search` or method `get_connected_cubes` from `Document` class
# or method `get_dependencies` from `Entity` class)
cubes = Dossier(conn, id=OBJECT_ID).get_connected_cubes()
cubes = Dossier(conn, id=OBJECT_ID).list_dependencies(
    project=PROJECT_ID, object_types=[ObjectSubTypes.OLAP_CUBE, ObjectSubTypes.SUPER_CUBE]
)
cubes = full_search(
    conn,
    project=PROJECT_ID,
    object_types=[ObjectSubTypes.OLAP_CUBE, ObjectSubTypes.SUPER_CUBE],
    used_by_object_id=OBJECT_ID,
    used_by_object_type=ObjectTypes.DOCUMENT_DEFINITION
)

# we can also list cubes that use given dossier
cubes_using_dossier = Dossier(conn, id=OBJECT_ID).list_dependents(
    project=PROJECT_ID, object_types=[ObjectSubTypes.OLAP_CUBE, ObjectSubTypes.SUPER_CUBE]
)

# get all objects that use (also recursive) metric with given id
objects = full_search(
    conn,
    project=PROJECT_ID,
    uses_object_id=OBJECT_ID,
    uses_object_type=ObjectTypes.METRIC,
    uses_recursive=True
)



from enum import auto, Enum
import logging
from pathlib import Path
from typing import Callable, List, Optional, Union

from mstrio import config
from mstrio.api.exceptions import VersionException
from mstrio.connection import Connection
from mstrio.object_management.migration.package import Package, PackageConfig, PackageImport
from mstrio.utils.helper import Dictable, exception_handler
from mstrio.utils.progress_bar_mixin import ProgressBarMixin
from mstrio.utils.wip import module_wip, WipLevels

module_wip(globals(), level=WipLevels.PREVIEW)

logger = logging.getLogger(__name__)


class MigrationStatus(Enum):
    NOT_STARTED = auto()
    STARTED = auto()
    CREATING_PACKAGE = auto()
    CREATING_PACKAGE_FAILED = auto()
    CREATING_PACKAGE_COMPLETED = auto()
    MIGRATION_IN_PROGRESS = auto()
    MIGRATION_FAILED = auto()
    MIGRATION_COMPLETED = auto()
    UNDO_STARTED = auto()
    UNDO_FAILED = auto()
    UNDO_COMPLETED = auto()


class Migration(Dictable, ProgressBarMixin):
    """A class encapsulating migration process from env A to env B.

    Raises VersionException if either of environments are running IServer
    version lower than 11.3.0200

    Raises AttributeError if parent directory of specified `save_path` is
    non-existent or points to another kind of file.

    Attributes:
        configuration(PackageConfig): a configuration of the migration
        save_path(str): a full path used when saving import package to a
        file. It is also used for saving undo package with the addition of
        `_undo` suffix. By default, uses ~cwd/migration.mmp and
        ~cwd/migration_undo.mmp.
        source_connection(Connection): A MicroStrategy connection object
          to the source env,
        target_connection(Connection): A MicroStrategy connection object
          to the target env,
        name(str): Name of the migration. Used for identification purposes for
            the convenience of the user,
        package(Package): a package that is migrated,
        package_import(PackageImport): a package import process object,
        package_binary(Binary): actual data that is migrated
        status(MigrationStatus): current status of migration,
        create_undo(bool): should a undo package be made
        undo_package(Binary): undo package binary for this import process.
          if `create_undo` is False, it is equal to None.

    Additional args:
        custom_package_path(str): a full path to an already existing .mpp file.
            This argument overrides `save_path`. Creating a `Migration` object
            with only `source_connection`, `target_connection` and
            `custom_package_path` allows for an easier export to the target by
            running `migrate_package()` on that `Migration` instance (instead of
            performing full migration with an import from the source first.
            See demos and examples for details.)."""
    _DELETE_NONE_VALUES_RECURSION = False

    def __init__(
        self,
        save_path: Optional[str] = f"{Path.cwd()}/migration.mmp",
        source_connection: Optional[Connection] = None,
        configuration: Optional[PackageConfig] = None,
        target_connection: Optional[Connection] = None,
        custom_package_path: Optional[str] = None,
        name: Optional[str] = None,
        create_undo: bool = True
    ):
        if not self._validate_envs_version(source_connection, target_connection):
            exception_handler(
                msg=(
                    "Environments must run IServer version 11.3.0200 or newer. "
                    "Please update your environments to use this feature."
                ),
                exception_type=VersionException
            )
        self.name = name
        self.source_connection = source_connection
        self.target_connection = target_connection
        self.configuration = configuration
        self.create_undo = create_undo
        self._status = MigrationStatus.NOT_STARTED

        self._check_file_path(save_path, "save_path")
        self.save_path = save_path

        self._package = None
        self._package_import = None
        self._undo_binary = None

        if custom_package_path and self._check_file_path(custom_package_path,
                                                         "custom_package_path"):
            self.save_path = custom_package_path
            filename, file_extension = self._decompose_file_path(custom_package_path)
            with open(f"{filename}{file_extension}", "rb") as f:
                self._package_binary = f.read()
        else:
            self._package_binary = None

    @staticmethod
    def _validate_envs_version(
        source_connection: Connection, target_connection: Connection
    ) -> bool:
        return (
            (source_connection is None or source_connection._iserver_version >= '11.3.0200')
            and (target_connection is None or target_connection._iserver_version >= '11.3.0200')
        )

    @staticmethod
    def _decompose_file_path(file_path: str) -> tuple:
        file_path = Path(file_path)
        filename, file_extension = file_path.parent / file_path.stem, file_path.suffix
        if file_extension == '':
            return filename, '.mmp'
        return filename, file_extension

    @staticmethod
    def _check_file_path(file_path: str, var_name: str):
        file_path = Path(file_path)

        if file_path.parent.is_dir() and not file_path.is_dir():
            return True
        else:
            exception_handler(
                msg=f"Invalid save path. Parent directory specified in `{var_name}` is incorrect.",
                exception_type=AttributeError
            )

    def perform_full_migration(self) -> bool:
        """Perform 'create_package()' and 'migrate_package()' using
        configuration provided when creating `Migration` object.
        """
        if not isinstance(self.source_connection, Connection) or self.source_connection is None:
            exception_handler(
                msg=(
                    "Migration object missing `source_connection`. "
                    "`perform_full_migration()` unavailable."
                ),
                exception_type=AttributeError
            )
        if not isinstance(self.target_connection, Connection) or self.target_connection is None:
            exception_handler(
                msg=(
                    "Migration object missing `target_connection`. "
                    "`perform_full_migration()` unavailable."
                ),
                exception_type=AttributeError
            )

        self._display_progress_bar(
            desc='Migration status: ',
            unit='Migration',
            total=4,
            bar_format='{desc} |{bar:50}| {percentage:3.0f}%'
        )

        if not self.create_package():
            return False

        if not self.migrate_package():
            return False

        self._close_progress_bar()
        return True

    def create_package(self) -> bool:
        """Performs import of the object described in the migration
        configuration from the source environment and save it in a file
        at location specified in `save_path` parameter.

        Raises AttributeError if `source_connection` is not specified.

        Raises FileExistsError if a package or undo package with the same name
        already exist at`save_path` location. """
        if not isinstance(self.source_connection, Connection) or self.source_connection is None:
            exception_handler(
                msg=("Migration object does not have `source_connection`. Import unavailable."),
                exception_type=AttributeError
            )

        filename, file_extension = self._decompose_file_path(self.save_path)
        """File existence checked here instead of using "xb" mode in context
        manager to avoid creation of package if it should not be saved"""
        if Path(f"{filename}{file_extension}").exists() or Path(f"{filename}_undo{file_extension}"
                                                                ).exists():
            exception_handler(
                msg=(
                    "Migration file / undo package with this name already exists."
                    "Please use other location / filename in `save_path` parameter."
                ),
                exception_type=FileExistsError
            )

        self.__private_status = MigrationStatus.CREATING_PACKAGE
        self._create_package_holder(self.source_connection)
        if not self._update_package_holder():
            self.__private_status = MigrationStatus.CREATING_PACKAGE_FAILED
            return False

        self._download_package_binary()
        self._delete_package_holder()
        self._save_package_binary_locally(filename, file_extension, self.package_binary)

        self.__private_status = MigrationStatus.CREATING_PACKAGE_COMPLETED
        return True

    def migrate_package(self, custom_package_path: Optional[str] = None, is_undo=False) -> bool:
        """Performs migration of already created package to the target
        environment. Import package will be loaded from `custom_package_path`.
        If `custom_package_path` not provided, the object previously acquired
        with the `create_package()` will be used.
        If `create_undo` parameter is set to True, package needed for undo
        process will be downloaded.

        Raises AttributeError if `target_connection` is not specified.
        """

        if not isinstance(self.target_connection, Connection) or self.target_connection is None:
            exception_handler(
                msg=("Migration object does not have `target_connection`. "
                     "Export unavailable."),
                exception_type=AttributeError
            )

        if custom_package_path and self._check_file_path(custom_package_path,
                                                         "custom_package_path"):
            self.save_path = custom_package_path
            filename, file_extension = self._decompose_file_path(custom_package_path)
            with open(f"{filename}{file_extension}", "rb") as f:
                return self._migrate_package(f.read(), is_undo=is_undo)
        return self._migrate_package(
            self._package_binary, custom_package_path=custom_package_path, is_undo=is_undo
        )

    def _migrate_package(
        self, binary: bytes, is_undo: bool = False, custom_package_path: Optional[str] = None
    ) -> bool:
        if binary is None:
            exception_handler(
                msg=(
                    "Import package is None. Run `create_package()` first, "
                    "or specify `custom_package_path`."
                ),
                exception_type=AttributeError
            )

        self.__private_status = MigrationStatus.MIGRATION_IN_PROGRESS
        self._create_package_holder(self.target_connection)
        self._upload_package_binary(binary)
        if not self._create_import(self.target_connection):
            self.__private_status = MigrationStatus.MIGRATION_FAILED
            return False

        if not is_undo and self.create_undo:
            self._download_undo_binary()
            file_path = custom_package_path if custom_package_path else self.save_path
            filename, file_extension = self._decompose_file_path(file_path)
            self._save_package_binary_locally(
                filename=f"{filename}_undo",
                file_extension=file_extension,
                _bytes=self.undo_binary
            )

        self._delete_package_holder()
        self._delete_import()

        self.__private_status = MigrationStatus.MIGRATION_COMPLETED
        return True

    def undo_migration(self):
        """Revert the migration using the package downloaded during
        `migrate_package()` or `perform_full_migration()`.

        Raises AttributeError if `undo_binary` is None.
        """
        self.__private_status = MigrationStatus.UNDO_STARTED
        if self.undo_binary:
            self._migrate_package(self._undo_binary, True)
        else:
            self.__private_status = MigrationStatus.UNDO_FAILED
            msg = (
                '`undo_binary` is None. Perform migration with `create_undo`'
                ' parameter set to True'
            )
            exception_handler(msg, exception_type=AttributeError)
        self.__private_status = MigrationStatus.UNDO_COMPLETED

    def _create_package_holder(self, conn: Connection):
        self._package = Package.create(conn, progress_bar=self._progress_bar is not None)

    def _update_package_holder(self):
        return self._package.update_config(self.configuration)

    def _download_package_binary(self):
        self._package_binary = self._package.download_package_binary(
            progress_bar=self._progress_bar is not None
        )

    def _upload_package_binary(self, binary: bytes):
        self._package.upload_package_binary(binary, progress_bar=self._progress_bar is not None)

    def _delete_package_holder(self):
        self._package.delete(force=True)
        self._package = None

    def _create_import(self, conn: Connection) -> bool:
        self._package_import = PackageImport.create(
            conn, self.package.id, self.create_undo, progress_bar=self._progress_bar is not None
        )
        if self._package_import is False:
            return False
        return True

    def _delete_import(self):
        self._package_import.delete(force=True)
        self._package_import = None

    def _download_undo_binary(self):
        self._undo_binary = self._package_import.download_undo_binary(
            progress_bar=self._progress_bar is not None
        )

    def _save_package_binary_locally(self, filename: str, file_extension: str, _bytes: bytes):
        with open(f"{filename}{file_extension}", "wb") as f:
            f.write(_bytes)
        logger.info(f'Package / package undo binary created at:{filename}{file_extension}')

    @property
    def package(self):
        return self._package

    @property
    def package_binary(self):
        return self._package_binary

    @property
    def package_import(self):
        return self._package_import

    @property
    def undo_binary(self):
        return self._undo_binary

    @property
    def status(self):
        return self._status

    @property
    def __private_status(self):
        return self._status

    @__private_status.setter
    def __private_status(self, var: MigrationStatus):
        self._update_progress_bar_if_needed(new_description=f'[{self.name}] Status: {var.name}')
        self._status = var


def bulk_full_migration(migrations: Union[Migration, List[Migration]], verbose: bool = False):
    """Run `perform_full_migration()` for each of the migrations provided.
    Args:
        migrations: migrations to be executed
        verbose: if True, information about each step will be printed
    """
    __run_func_for_all_migrations(Migration.perform_full_migration, migrations, verbose)


def bulk_migrate_package(migrations: Union[Migration, List[Migration]], verbose: bool = False):
    """Run `migrate_package()` for each of the migrations provided.
    Args:
        migrations: migrations to be executed
        verbose: if True, information about each step will be printed
    """
    __run_func_for_all_migrations(Migration.migrate_package, migrations, verbose)


def __run_func_for_all_migrations(
    func: Callable,
    migrations: Union[Migration, List[Migration]],
    verbose: bool = False,
    *args,
    **kwargs
):
    if type(migrations) is Migration:
        migrations = [migrations]
    original_verbose = config.verbose
    config.verbose = verbose
    for migration in migrations:
        func(migration, *args, **kwargs)
    config.verbose = original_verbose



from enum import Enum
from itertools import chain
from typing import Tuple, Union


class AutoName(Enum):

    def _generate_next_value_(name, start, count, last_values):
        return name.lower()

    @classmethod
    def has_value(cls, value):
        return value in set(item.value for item in cls)


class AutoUpperName(Enum):

    def _generate_next_value_(name, start, count, last_values):
        return name


def __get_enum_helper(obj, enum: Union[Enum, Tuple[Enum]] = Enum,
                      get_value: bool = False) -> Union[str, Enum, None]:
    """Helper function for `get_enum` and `get_enum_val`."""
    if obj is None:
        return obj
    elif isinstance(obj, Enum) and isinstance(obj, enum):
        obj = obj.value if get_value else obj
    elif isinstance(obj, (str, int)):
        validate_enum_value(obj, enum)
        obj = obj if get_value else enum(obj)
    else:
        raise TypeError(f"Incorrect type. Value should be of type: {enum}.")

    return obj


def get_enum(obj, enum: Union[Enum, Tuple[Enum]] = Enum) -> Union[Enum, None]:
    """Safely get enum from enum or str."""
    return __get_enum_helper(obj, enum)


def get_enum_val(obj, enum: Union[Enum, Tuple[Enum]] = Enum) -> Union[str, None]:
    """Safely extract value from enum or str."""
    return __get_enum_helper(obj, enum, True)


def validate_enum_value(obj, enum):
    """Validate provided value. If not correct,
    error message with possible options will be displayed.
    """
    from mstrio.utils.helper import exception_handler
    possible_values = [[e.value
                        for e in item]
                       for item in enum] if isinstance(enum, tuple) else [e.value for e in enum]
    err_msg = f"Incorrect enum value '{obj}'. Possible values are {possible_values}"
    if obj not in list(chain(possible_values)):
        exception_handler(err_msg, exception_type=ValueError)



# flake8: noqa
# isort: off
from .dimensionality import Dimensionality, DimensionalityUnit
from .metric_format import FormatProperty, MetricFormat
from .metric import DefaultSubtotals, FormatProperty, list_metrics, Metric, MetricFormat, Threshold
# isort: on



# flake8: noqa

# isort: off
from .dbms import Dbms, list_available_dbms
# isort: on
from .database_connections import DatabaseConnections
from .datasource_connection import (
    CharEncoding, DatasourceConnection, DriverType, ExecutionMode, list_datasource_connections
)
from .datasource_instance import (
    DatasourceInstance,
    DatasourceType,
    list_connected_datasource_instances,
    list_datasource_instances
)
from .datasource_login import DatasourceLogin, list_datasource_logins
from .datasource_map import DatasourceMap, list_datasource_mappings, Locale



"""This is the demo script to show how to manage user hierachies.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.modeling.schema import (
    ElementDisplayOption,
    HierarchyAttribute,
    HierarchyRelationship,
    SchemaManagement,
    SchemaObjectReference,
    SchemaUpdateType,
    ObjectSubType,
    UserHierarchy,
    UserHierarchySubType,
    list_user_hierarchies
)

from mstrio.connection import get_connection

PROJECT_NAME = '<project_name>'  # Project to connect to
OBJECT_ID = '<object_id>'  # Object ID for SchemaObjectReference
ATTRIBUTE_NAME_1 = '<attribute_name_1>'
ATTRIBUTE_NAME_2 = '<attribute_name_2>'
ATTRIBUTE_NAME_3 = '<attribute_name_3>'
ATTRIBUTE_NAME_4 = '<attribute_name_4>'
# see modeling/schema/user_hierarchy/user_hierarchy.py -
# - ElementDisplayOption class for available options
ELEMENT_DISPLAY_OPTION = '<element_display_option>'
USER_HIERARCHY_NAME = '<user_hierarchy_name>'
USER_HIERARCHY_DESCRIPTION = 'user_hierarchy_description>'
FOLDER_ID = '<folder_id>'
SUBTYPE = '<sub_type>'

conn = get_connection(workstationData, project_name=PROJECT_NAME)
# get a list of user hierarchies
user_hierarchies = list_user_hierarchies(conn)

# create attributes to be added to a user hierarchy
# an attribute can be either an object (HierarchyAttribute) or a dict
attribute1 = HierarchyAttribute(
    object_id=OBJECT_ID,
    entry_point=True,
    name=ATTRIBUTE_NAME_1,
    # see modeling/schema/user_hierarchy/user_hierarchy.py for available options
    element_display_option=ElementDisplayOption.ALL_ELEMENTS,
)

attribute2 = {
    "objectId": OBJECT_ID,
    "entryPoint": True,
    "name": ATTRIBUTE_NAME_2,
    "elementDisplayOption": ELEMENT_DISPLAY_OPTION
}

# create SchemaObjectReference objects to be used
# for relationship creation
sor1 = SchemaObjectReference(
    object_id=OBJECT_ID,
    sub_type=ObjectSubType.ATTRIBUTE  # see modeling/schema/helpers.py - ObjectSubType class
    # for available options
)

sor2 = SchemaObjectReference(
    object_id=OBJECT_ID,
    sub_type=ObjectSubType.ATTRIBUTE  # see modeling/schema/helpers.py - ObjectSubType class
    # for available options
)

# create a hierarhcy relationship between two attributes
relationship = HierarchyRelationship(parent=sor1, child=sor2)

# create a user hierarchy with subtype `DIMENSION_USER_HIERARCHY`
# and previously created atributes and relationship
new_user_hierarchy = UserHierarchy.create(
    connection=conn,
    name=USER_HIERARCHY_NAME,
    # see modeling/schema/user_hierarchy/user_hierarchy.py for available options
    sub_type=UserHierarchySubType.DIMENSION_USER_HIERARCHY,
    attributes=[attribute1, attribute2],
    relationships=[relationship],
    destination_folder_id=FOLDER_ID,
)

# Any changes to a schema objects must be followed by schema_reload
# in order to use them in reports, dossiers and so on
schema_manager = SchemaManagement(connection=conn, project_id=conn.project_id)
task = schema_manager.reload(update_types=[SchemaUpdateType.LOGICAL_SIZE])

# alter a user hierarchy, change it's name, description
# and subtype
new_user_hierarchy.alter(
    name=USER_HIERARCHY_NAME,
    description=USER_HIERARCHY_DESCRIPTION,
    # see modeling/schema/user_hierarchy/user_hierarchy.py for available options
    sub_type=UserHierarchySubType.DIMENSION_USER,
    use_as_drill_hierarchy=False
)

# define additional attributes
attribute3 = HierarchyAttribute(
    object_id=OBJECT_ID,
    entry_point=True,
    name=ATTRIBUTE_NAME_3,
    element_display_option=ElementDisplayOption.ALL_ELEMENTS,
)

attribute4 = HierarchyAttribute(
    object_id=OBJECT_ID,
    entry_point=False,
    name=ATTRIBUTE_NAME_4,
    element_display_option=ElementDisplayOption.ALL_ELEMENTS,
    limit=50
)

# add new attributes to a hierarchy
new_user_hierarchy.add_attribute(attribute3)
new_user_hierarchy.add_attribute(attribute4)

# define another relationship (it may be a dict)
relationship = {
    'parent': {
        'objectId': OBJECT_ID,
        'subType': ObjectSubType.ATTRIBUTE  # see modeling/schema/helpers.py for available values
    },
    'child': {
        'objectId': OBJECT_ID,
        'subType': ObjectSubType.ATTRIBUTE  # see modeling/schema/helpers.py for available values
    }
}

# add new relationship to a hierarchy
new_user_hierarchy.add_relationship(relationship)

# remove a relationship from a hierarchy
new_user_hierarchy.remove_relationship(relationship)

# remove an attribute from a hierarchy
new_user_hierarchy.remove_attribute(attribute3)

# get a user hierarchy by ID. User hierarchy can be also found by its name.
user_hierarchy = UserHierarchy(conn, id=new_user_hierarchy.id)
user_hierarchy_by_name = UserHierarchy(conn, name=USER_HIERARCHY_NAME)

# list properties for user hierarchy
user_hierarchy.list_properties()

# Delete a user_hierarchy.
# When argument `force` is set to `False` (default value),
# then deletion must be confirmed by selecting appropriate prompt value.
user_hierarchy.delete(force=True)



"""This is the demo script to show how administrator can manage projects
and its settings.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.server import compare_project_settings, Environment, Project
from mstrio.connection import get_connection

PROJECT_NAME = '<Project_name>'  # Insert name of project to interact with
NODE_NAME = '<Node_name>'  # Insert name of node, that has a project, to load or unload it

PROJECT_DESCRIPTION = '<Description_of_project>'  # Insert description of newly created project

# These strings are used for project settings comparison
PROJECT_1_NAME = '<Project_name>'
PROJECT_2_NAME = '<Project_name>'
CSV_FILE_EXPORT_IMPORT = 'path/to/file.csv'  # Name of file to/from which settings are imported

conn = get_connection(workstationData, project_name=PROJECT_NAME)
env = Environment(connection=conn)

# get list of all projects or those just loaded
all_projects = env.list_projects()
loaded_projects = env.list_loaded_projects()

# get project with a given name
project = Project(connection=conn, name=PROJECT_NAME)
# idle/resume a project on a given node(s) (name(s) of existing node(s)
# should be used)
project.idle(on_nodes=[NODE_NAME])
project.resume(on_nodes=NODE_NAME)
# unload/load a project from/to a given node(s) (name(s) of existing
# node(s) should be used)
project.unload(on_nodes=NODE_NAME)
project.load(on_nodes=NODE_NAME)

# get settings of a project as a dataframe
project_settings_df = project.settings.to_dataframe()

# create a project and store it in variable to have immediate access to it
new_project = env.create_project(name=PROJECT_NAME, description=PROJECT_DESCRIPTION)

# change description of a newly created project
new_project.alter(description=PROJECT_DESCRIPTION)

# compare settings of 2 projects (only differences will be stored into
# a DataFrame)
project1 = Project(connection=conn, name=PROJECT_1_NAME)
project2 = Project(connection=conn, name=PROJECT_2_NAME)
df_cmp = compare_project_settings(projects=[project, project1, project2], show_diff_only=True)

# save/load settings of a project to/from a file (format can be 'csv',
# 'json' or 'pickle')
project.settings.to_csv(name=CSV_FILE_EXPORT_IMPORT)
project.settings.import_from(file=CSV_FILE_EXPORT_IMPORT)

# change a setting of a project
project.settings.cubeIndexGrowthUpperBound = 501
project.settings.update()



"""This is the demo script to show how to manage transmitters.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import get_connection
from mstrio.distribution_services import (
    EmailTransmitterProperties,
    list_transmitters,
    RecipientFieldType,
    Transmitter,
    TransmitterDeliveryType
)

PROJECT_NAME = '<project_name>'  # Project to connect to
SENDER_DISPLAY_NAME = '<sender_display_name>'  # name of a sender
SENDER_EMAIL_ADDRESS = '<sender_email_address>'  # address of a sender
REPLY_TO_DISPLAY_NAME = '<reply_to_display_name>'  # name for destination for replies
REPLY_TO_EMAIL_ADDRESS = '<reply_to_email_address>'  # address for destination for replies
TRANSMITTER_ID = '<transmitter_id>'
TRANSMITTER_NAME = '<transmitter_name>'
TRANSMITTER_DESCRIPTION = '<transmitter_description>'
NOTIFICATION_EMAIL_ADDRESS = '<notification_email_address>'

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# create email transmitter properties object
# which is used when creating transmitter
etp = EmailTransmitterProperties(
    sender_display_name=SENDER_DISPLAY_NAME,
    sender_email_address=SENDER_EMAIL_ADDRESS,
    reply_to_display_name=REPLY_TO_DISPLAY_NAME,
    reply_to_email_address=REPLY_TO_EMAIL_ADDRESS,
    # see distribution_services/transmitter/transmitter.py - RecipientFieldType
    # for available values
    recipient_field_type=RecipientFieldType.TO,
    save_message_to_file=False,
    send_message_to_file=False,
    send_message_via_smtp=False,
    save_file_path=None,
    notify_on_success=False,
    notify_on_failure=False,
    notification_email_address=None,
)

# create a transmitter with delivery type as `email` (when type is `email` then
# it is mandatory to provide `email_transmitter_properties`). Delivery type
# can be also passed as a string and `email_transmitter_properties` can be
# passed as a dictionary
new_t = Transmitter.create(
    connection=conn,
    name=TRANSMITTER_NAME,
    delivery_type=TransmitterDeliveryType.EMAIL,
    description=TRANSMITTER_DESCRIPTION,
    email_transmitter_properties=etp,
)

# alter email transmitter and its properties
etp.notify_on_success = True
etp.notify_on_failure = True
etp.notification_email_address = NOTIFICATION_EMAIL_ADDRESS
etp.save_message_to_file = True
new_t.alter(
    name=TRANSMITTER_NAME,
    description=TRANSMITTER_DESCRIPTION,
    email_transmitter_properties=etp,
)

# create file transmitter
new_file_t = Transmitter.create(
    connection=conn,
    name=TRANSMITTER_NAME,
    delivery_type=TransmitterDeliveryType.FILE,
    description=TRANSMITTER_DESCRIPTION,
)

# alter file transmitter
new_file_t.alter(name=TRANSMITTER_NAME, description=TRANSMITTER_DESCRIPTION)

# create ipad transmitter
new_ipad_t = Transmitter.create(
    connection=conn,
    name=TRANSMITTER_NAME,
    delivery_type=TransmitterDeliveryType.IPAD,
    description=TRANSMITTER_DESCRIPTION,
)

# alter ipad transmitter
new_ipad_t.alter(name=TRANSMITTER_NAME, description=TRANSMITTER_DESCRIPTION)

# create iphone transmitter
new_iphone_t = Transmitter.create(
    connection=conn,
    name=TRANSMITTER_NAME,
    delivery_type=TransmitterDeliveryType.IPHONE,
    description=TRANSMITTER_DESCRIPTION,
)

# alter iphone transmitter
new_iphone_t.alter(name=TRANSMITTER_NAME, description=TRANSMITTER_DESCRIPTION)

# create print transmitter
new_print_t = Transmitter.create(
    connection=conn,
    name=TRANSMITTER_NAME,
    delivery_type=TransmitterDeliveryType.PRINT,
    description=TRANSMITTER_DESCRIPTION,
)

# alter print transmitter
new_print_t.alter(name=TRANSMITTER_NAME, description=TRANSMITTER_DESCRIPTION)

# get list of transmitters
transmitters = list_transmitters(conn)

# get transmitter by ID. Transmitter can be also found by its name.
t = Transmitter(conn, id=TRANSMITTER_ID)
t_by_name = Transmitter(conn, name=TRANSMITTER_NAME)

# list properties for transmitter
t.list_properties()

# Delete transmitter. When argument `force` is set to `False` (default value),
# then deletion must be confirmed by selecting appropriate prompt value.
t.delete(force=True)



from datetime import datetime, timezone
from enum import auto
import logging
from typing import List, Optional, Union

from mstrio import config
from mstrio.api import objects, schedules
from mstrio.connection import Connection
from mstrio.distribution_services.event import Event
from mstrio.distribution_services.schedule import ScheduleEnums, ScheduleTime
from mstrio.users_and_groups.user import User
from mstrio.utils import helper
from mstrio.utils.entity import DeleteMixin, Entity, ObjectTypes
from mstrio.utils.enum_helper import AutoName, get_enum_val
from mstrio.utils.time_helper import DatetimeFormats, map_datetime_to_str, map_str_to_datetime
from mstrio.utils.version_helper import method_version_handler

logger = logging.getLogger(__name__)


@method_version_handler('11.3.0000')
def list_schedules(
    connection: Connection, to_dictionary: bool = False, limit: int = None, **filters
) -> Union[List["Schedule"], List[dict]]:
    """List schedule objects or schedule dictionaries. Optionally filter list.

    Args:
        connection(object): MicroStrategy connection object returned by
            'connection.Connection()'
        to_dictionary(bool, optional): if True, return Schedules as
            list of dicts
        limit(int, optional): maximum number of schedules returned.
        **filters: Available filter parameters:['name':,
                                                'id',
                                                'description',
                                                'schedule_type',
                                                'start_date',
                                                'expired']
    Returns:
        Union[List["Schedule"], List[dict]]: [description]
    """

    objects = helper.fetch_objects(
        connection=connection,
        api=schedules.list_schedules,
        limit=limit,
        filters=filters,
        dict_unpack_value='schedules'
    )

    if to_dictionary:
        return objects
    else:
        return [Schedule.from_dict(source=obj, connection=connection) for obj in objects]


class Schedule(Entity, DeleteMixin):
    """Class representation of MicroStrategy Schedule object.

    Attributes:
        connection: A MicroStrategy connection object
        name: Schedule name
        id: Schedule ID
        description: Schedule description
        schedule_type: Schedule type
        schedule_next_delivery: Schedule next delivery date
        time(ScheduleTime): Details of time-based schedule
        event(Event): Details of event-based schedule
    """

    class ScheduleType(AutoName):
        """Class representation of a type of a Microstrategy Schedule."""
        EVENT_BASED = auto()
        TIME_BASED = auto()
        NONE = None

    _DELETE_NONE_VALUES_RECURSION = False
    _OBJECT_TYPE = ObjectTypes.SCHEDULE_TRIGGER

    _API_GETTERS: dict = {
        (
            'id',
            'name',
            'description',
            'expired',
            'schedule_type',
            'schedule_next_delivery',
            'start_date',
            'stop_date',
            'time',
            'event'
        ): schedules.get_schedule,
        (
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl'
        ): objects.get_object_info
    }

    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'owner': User.from_dict,
        'schedule_type': ScheduleType,
        'time': ScheduleTime.from_dict,
        'event': Event.from_dict,
        'schedule_next_delivery': DatetimeFormats.YMDHMS,
        'start_date': DatetimeFormats.DATE,
        'stop_date': DatetimeFormats.DATE,
    }
    _API_PATCH: dict = {
        ('abbreviation'): (objects.update_object, 'partial_put'),
        ('name', 'description', 'schedule_type', 'start_date', 'stop_date', 'time',
         'event'): (schedules.update_schedule, 'put')
    }

    _PATCH_PATH_TYPES = {
        'name': str,
        'description': str,
        'abbreviation': str,
        'schedule_type': str,
        'start_date': str,
        'stop_date': str,
        'time': dict,
        'event': dict,
    }

    @method_version_handler('11.3.0000')
    def __init__(self, connection: Connection, id: str = None, name: str = None) -> None:
        """Initialize the Schedule object, populates it with I-Server data.
        Specify either `id` or `name`. When `id` is provided (not `None`),
        `name` is omitted.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`.
            id: Schedule ID
            name: Schedule name
        Raises:
            AttributeError: when id is None and name is None.
            ValueError: when there is no schedule with a given name on the
            server.
        """

        if id is None and name is None:
            raise AttributeError(
                "Please specify either 'name' or 'id' parameter in the constructor."
            )
        if id is None:
            objects_info = list_schedules(connection=connection, name=name, to_dictionary=True)
            if objects_info:
                object_info, object_info["connection"] = objects_info[0], connection
                self._init_variables(**object_info)
            else:
                raise ValueError(f"There is no schedule with the given name: '{name}'")
        else:
            super().__init__(connection, id, name=name)

    @method_version_handler('11.3.0000')
    def _init_variables(self, **kwargs):
        """
        Add Schedule attributes to _AVAILABLE_ATTRIBUTES map.
        Set attributes inherited from Entity object.
        Set attributes related to schedule details.

        Args:
            **kwargs: dict with information about schedule attributes.
        Returns:
            None
        """
        super()._init_variables(**kwargs)
        self.schedule_type = self.ScheduleType(kwargs.get('schedule_type'))
        self.time = ScheduleTime.from_dict(
            kwargs.get('time')
        ) if (self.schedule_type == self.ScheduleType.TIME_BASED and kwargs.get('time')) else None
        self.event = Event.from_dict(kwargs.get('event'), connection=self._connection) if (
            self.schedule_type == self.ScheduleType.EVENT_BASED and kwargs.get('event')
        ) else None
        self._expired = kwargs.get('expired')
        self._schedule_next_delivery = map_str_to_datetime(
            "schedule_next_delivery", kwargs.get("schedule_next_delivery"), self._FROM_DICT_MAP
        )
        self.start_date = map_str_to_datetime(
            "start_date", kwargs.get("start_date"), self._FROM_DICT_MAP
        )

    @method_version_handler('11.3.0000')
    def enable(self, stop_date: Union[str, datetime]) -> bool:
        """Enables schedule and sets stop date

        Args:
            stop_date: stop date provided either as a datetime or
            as a string in yyyy-MM-dd format
        Returns:
            Returns `True` if enabled properly, else `False`.
        """
        self._alter_properties(
            stop_date=map_str_to_datetime('stop_date', stop_date, self._FROM_DICT_MAP)
        )

        if config.verbose and not self.expired:
            logger.info(
                f'Schedule \'{self.name}\' with ID {self.id} has been enabled, '
                f'until {self.stop_date}.'
            )
            return True
        elif config.verbose:
            logger.info(f'Schedule \'{self.name}\' with ID {self.id} has NOT been enabled.')
            return False

    @method_version_handler('11.3.0000')
    def disable(self, stop_date: Union[str, datetime] = None) -> bool:
        """ Disable the schedule. Optional `stop_date` sets the date when
            the schedule should be disabled.

        Args:
            stop_date: stop date provided either as a datetime or
            as a string in yyyy-MM-dd format
        Returns:
            Returns `True` if disabled properly. It does not mean that schedule
            is already expired, as it can take up to one day.
            If operation failed, return `False`.
        """
        stop_date = (
            datetime.now(timezone.utc) if stop_date is None else
            map_str_to_datetime('stop_date', stop_date, self._FROM_DICT_MAP)
        )
        self._alter_properties(stop_date=stop_date)

        if config.verbose and self.expired:
            logger.info(f'Schedule \'{self.name}\' with ID {self.id} has been disabled.')
            return True
        elif config.verbose and self.stop_date.date() == stop_date.date():
            logger.info(
                f"Schedule '{self.name}' with ID '{self.id}' has been set for disabling. "
                f"Depending on the schedule configuration (`event`, `time` and `stop_date`),"
                f" it will be disabled by day after '{self.stop_date.date()}'."
            )
            return True
        else:
            logger.info(f"Schedule '{self.name}' with ID '{self.id}' has NOT been disabled.")
            return False

    @method_version_handler('11.3.0000')
    def list_properties(self):
        """List all properties of the object."""
        attributes = {key: self.__dict__[key] for key in self.__dict__ if not key.startswith('_')}
        attributes = {
            **attributes,
            'expired': self._expired,
            'schedule_next_delivery': self._schedule_next_delivery,
            'id': self.id,
            'type': self.type,
            'subtype': self.subtype,
            'ext_type': self.ext_type,
            'date_created': self.date_created,
            'date_modified': self.date_modified,
            'version': self.version,
            'owner': self.owner,
            'icon_path': self.icon_path,
            'view_media': self.view_media,
            'ancestors': self.ancestors,
            'certified_info': self.certified_info,
            'acg': self.acg,
            'acl': self.acl,
        }

        return {
            key: attributes[key]
            for key in sorted(attributes, key=helper.sort_object_properties)
        }

    @classmethod
    @method_version_handler('11.3.0000')
    def create(
        cls,
        connection: Connection,
        name: str,
        schedule_type: Union[ScheduleType, str],
        start_date: Union[str, datetime],
        description: Optional[str] = None,
        stop_date: Optional[Union[str, datetime]] = None,
        event_id: Optional[str] = None,
        time: Optional[ScheduleTime] = None,
        recurrence_pattern: Optional[Union[ScheduleEnums.RecurrencePattern, str]] = None,
        execution_pattern: Optional[Union[ScheduleEnums.ExecutionPattern, str]] = None,
        execution_time: Optional[str] = None,
        start_time: Optional[str] = None,
        stop_time: Optional[str] = None,
        execution_repeat_interval: Optional[int] = None,
        daily_pattern: Optional[Union[ScheduleEnums.DailyPattern, str]] = None,
        repeat_interval: Optional[int] = None,
        days_of_week: Optional[Union[List[ScheduleEnums.DaysOfWeek], List[str]]] = None,
        day: Optional[int] = None,
        month: Optional[int] = None,
        week_offset: Optional[Union[ScheduleEnums.WeekOffset, str]] = None,
        day_of_week: Optional[Union[ScheduleEnums.DaysOfWeek, str]] = None,
        weekday_off_set: Optional[str] = None,
        days_of_month: Optional[List[str]] = None,
        monthly_pattern: Optional[Union[ScheduleEnums.MonthlyPattern, str]] = None,
        yearly_pattern: Optional[Union[ScheduleEnums.YearlyPattern, str]] = None
    ):
        """Create a Schedule using provided parameters as data.

        Args:
            recurrence_pattern (ScheduleEnums.RecurrencePattern, optional):
                The recurrence pattern of the schedule. Possible values are
                DAILY, WEEKLY, MONTHLY, YEARLY. Defaults to None.
            execution_pattern (ScheduleEnums.ExecutionPattern, optional):
                The execution pattern of the schedule. Possible values are ONCE,
                REPEAT. Defaults to None.
            execution_time (str, optional):
                The execution time of the execution day, if execution_pattern
                is ONCE. Format should be HH:mm:ss. Defaults to None.
            start_time (str, optional):
                The start time of the execution day, if execution_pattern is
                REPEAT. Format should be HH:mm:ss. Defaults to None.
            stop_time (str, optional):
                The stop time of the execution day, if execution_pattern is
                REPEAT.Format should be HH:mm:ss. Defaults to None.
            execution_repeat_interval (int, optional):
                The repeat interval of minutes of the execution day, if
                execution_pattern is REPEAT. Defaults to None.
            daily_pattern (ScheduleEnums.DailyPattern, optional):
                The daily recurrence pattern of the schedule. Possible values
                are DAY, WEEKDAY. Defaults to None.
            repeat_interval (int, optional):
                The repeat interval of days of daily schedule, if daily_pattern
                is DAY. Defaults to None.
            day (int, optional):
                The day in month of monthly schedule, if monthly_pattern is DAY
                or, The day in month of yearly schedule, if yearly_pattern is
                DAY. Defaults to None.
            month (int, optional):
                The month in year of yearly schedule. Defaults to None.
            week_offset (ScheduleEnums.WeekOffset, optional):
                The week offset in month of monthly schedule, if monthly_pattern
                is DAY_OF_WEEK or, The week offset in year of yearly schedule,
                if yearly_pattern is DAY_OF_WEEK. Possible values are FIRST,
                SECOND, THIRD, FOURTH, LAST. Defaults to None.
            day_of_week (ScheduleEnums.DaysOfWeek, optional):
                The days of week of weekly schedule or, The day of week in month
                of monthly schedule, if monthly_pattern is DAY_OF_WEEK or, The
                day of week in year of yearly schedule, if yearly_pattern is
                DAY_OF_WEEK. Possible values are: MONDAY, TUESDAY, WEDNESDAY,
                THURSDAY, FRIDAY, SATURDAY, SUNDAY. Defaults to None.
            weekday_offset (ScheduleEnums.WeekdayOffset, optional):
                The weekday offset in month of monthly schedule, if
                monthly_pattern is WEEKDAY. Defaults to None.
            days_of_month (List[str], optional):
                The days of month of monthly schedule, if monthly_pattern is
                DAYS_OF_MONTH. Must be provided as a list of one or more
                stringified digits (from '1' to '31'). Defaults to None.
            monthly_pattern (ScheduleEnums.MonthlyPattern, optional):
                The monthly recurrence pattern of the schedule. Possible values
                are: DAY, DAY_OF_WEEK, WEEKDAY, LAST_DAY, DAYS_OF_MONTH.
                Defaults to None.
            yearly_pattern (ScheduleEnums.YearlyPattern, optional):
                The yearly recurrence pattern of the schedule. Possible values
                are DAY, DAY_OF_WEEK. Defaults to None.
        Returns:
            Schedule object with provided parameters.
        """
        time_kwargs = {
            key: val
            for key,
            val in locals().items()
            if val is not None and key not in [
                'event_id',
                'connection',
                'description',
                'name',
                'schedule_type',
                'start_date',
                'stop_date',
                'time',
                'self',
                'cls'
            ]
        }
        # Event based or Time based logic
        if schedule_type == cls.ScheduleType.EVENT_BASED:
            execution_details = {'type': 'event', 'content': {'id': event_id}}
        elif schedule_type == cls.ScheduleType.TIME_BASED:
            if time is None:
                time = ScheduleTime.from_details(**time_kwargs)
            execution_details = {'type': 'time', 'content': time.to_dict()}

        # Datetime dates to string format conversion
        start_date = map_datetime_to_str(
            name='start_date', date=start_date, string_to_date_map=cls._FROM_DICT_MAP
        )
        stop_date = map_datetime_to_str(
            name='stop_date', date=stop_date, string_to_date_map=cls._FROM_DICT_MAP
        )

        # Create body and send request
        body = {
            'name': name,
            'description': description,
            'schedule_type': get_enum_val(schedule_type, cls.ScheduleType),
            'start_date': start_date,
            'stop_date': stop_date,
            execution_details['type']: execution_details['content']
        }
        body = helper.delete_none_values(body, recursion=True)
        body = helper.snake_to_camel(body)
        # Response is already unpacked in wrapper
        response = schedules.create_schedule(connection, body)
        response = response.json()
        if config.verbose:
            logger.info(f"Created schedule '{name}' with ID: {response['id']}")
        return Schedule.from_dict(source=response, connection=connection)

    @method_version_handler('11.3.0000')
    def alter(
        self,
        name: str = None,
        description: str = None,
        start_date: Union[str, datetime] = None,
        stop_date: Union[str, datetime] = None,
        event: Event = None,
        event_id: str = None,
        time: ScheduleTime = None,
        recurrence_pattern: ScheduleEnums.RecurrencePattern = None,
        execution_pattern: ScheduleEnums.ExecutionPattern = None,
        execution_time: str = None,
        start_time: str = None,
        stop_time: str = None,
        execution_repeat_interval: int = None,
        daily_pattern: ScheduleEnums.DailyPattern = None,
        repeat_interval: int = None,
        days_of_week: List[ScheduleEnums.DaysOfWeek] = None,
        day: int = None,
        month: int = None,
        week_offset: ScheduleEnums.WeekOffset = None,
        day_of_week: ScheduleEnums.DaysOfWeek = None,
        weekday_offset: str = None,
        days_of_month: List[str] = None,
        monthly_pattern: ScheduleEnums.MonthlyPattern = None,
        yearly_pattern: ScheduleEnums.YearlyPattern = None
    ) -> None:
        """Alter Schedule properties.

        Args:
            recurrence_pattern (ScheduleEnums.RecurrencePattern, optional):
                The recurrence pattern of the schedule. Possible values are
                DAILY, WEEKLY, MONTHLY, YEARLY. Defaults to None.
            execution_pattern (ScheduleEnums.ExecutionPattern, optional):
                The execution pattern of the schedule. Possible values are ONCE,
                REPEAT. Defaults to None.
            execution_time (str, optional):
                The execution time of the execution day, if execution_pattern is
                ONCE. Format should be HH:mm:ss. Defaults to None.
            start_time (str, optional):
                The start time of the execution day, if execution_pattern is
                REPEAT. Format should be HH:mm:ss. Defaults to None.
            stop_time (str, optional):
                The stop time of the execution day, if execution_pattern is
                REPEAT. Format should be HH:mm:ss. Defaults to None.
            execution_repeat_interval (int, optional):
                The repeat interval of minutes of the execution day, if
                execution_pattern is REPEAT. Defaults to None.
            daily_pattern (ScheduleEnums.DailyPattern, optional):
                The daily recurrence pattern of the schedule. Possible values
                are DAY, WEEKDAY. Defaults to None.
            repeat_interval (int, optional):
                The repeat interval of days of daily schedule, if daily_pattern
                is DAY. Defaults to None.
            day (int, optional):
                The day in month of monthly schedule, if monthly_pattern is DAY
                or, The day in month of yearly schedule, if yearly_pattern is
                DAY. Defaults to None.
            month (int, optional):
                The month in year of yearly schedule. Defaults to None.
            week_offset (ScheduleEnums.WeekOffset, optional):
                The week offset in month of monthly schedule, if monthly_pattern
                is DAY_OF_WEEK or, The week offset in year of yearly schedule,
                if yearly_pattern is DAY_OF_WEEK. Possible values are FIRST,
                SECOND, THIRD, FOURTH, LAST. Defaults to None.
            day_of_week (ScheduleEnums.DaysOfWeek, optional):
                The days of week of weekly schedule or,
                The day of week in month of monthly schedule, if monthly_pattern
                is DAY_OF_WEEK or, The day of week in year of yearly schedule,
                if yearly_pattern is DAY_OF_WEEK. Possible values are: MONDAY,
                TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY, SUNDAY.
                Defaults to None.
            weekday_offset (ScheduleEnums.WeekdayOffset, optional):
                The weekday offset in month of monthly schedule, if
                monthly_pattern is WEEKDAY. Defaults to None.
            days_of_month (List[str], optional):
                The days of month of monthly schedule, if monthly_pattern is
                DAYS_OF_MONTH. Must be provided as a list of one or more
                stringified digits (from '1' to '31'). Defaults to None.
            monthly_pattern (ScheduleEnums.MonthlyPattern, optional):
                The monthly recurrence pattern of the schedule. Possible values
                are: DAY, DAY_OF_WEEK, WEEKDAY, LAST_DAY, DAYS_OF_MONTH.
                Defaults to None.
            yearly_pattern (ScheduleEnums.YearlyPattern, optional):
                The yearly recurrence pattern of the schedule. Possible values
                are DAY, DAY_OF_WEEK. Defaults to None.
        Returns:
            None
        """

        properties = {}

        # Event based or Time based logic
        if time is not None:
            properties['schedule_type'] = self.ScheduleType.TIME_BASED
            properties['time'] = time
        elif event is not None:
            properties['schedule_type'] = self.ScheduleType.EVENT_BASED
            properties['event'] = event
        elif event_id is not None:
            properties['schedule_type'] = self.ScheduleType.EVENT_BASED
            properties['event'] = Event(connection=self._connection, id=event_id)
        elif self.schedule_type == self.ScheduleType.TIME_BASED:
            properties['schedule_type'] = self.ScheduleType.TIME_BASED
            self.time.update_properties(
                recurrence_pattern,
                execution_pattern,
                execution_time,
                start_time,
                stop_time,
                execution_repeat_interval,
                daily_pattern,
                repeat_interval,
                days_of_week,
                day,
                month,
                week_offset,
                day_of_week,
                weekday_offset,
                days_of_month,
                monthly_pattern,
                yearly_pattern
            )
            properties['time'] = self.time

        if name:
            properties['name'] = name
        if description:
            properties['description'] = description
        if start_date:
            properties['start_date'] = map_str_to_datetime(
                'start_date', start_date, self._FROM_DICT_MAP
            )
        if stop_date:
            properties['stop_date'] = map_str_to_datetime(
                'stop_date', stop_date, self._FROM_DICT_MAP
            )

        self._alter_properties(**properties)

        if self.schedule_type == self.ScheduleType.EVENT_BASED and 'time' in self.__dir__():
            delattr(self, 'time')
        elif self.schedule_type == self.ScheduleType.TIME_BASED and 'event' in self.__dir__():
            delattr(self, 'event')

    @method_version_handler('11.3.0000')
    def delete(self) -> bool:
        """Delete the schedule.

        Returns:
            bool: True if deletion was successful else False.
        """
        # When the server behavior changes, and deleting last schedule
        # from subscription will no longer delete subscription,
        # move 'force' param to method definition
        force = False
        self._delete_confirm_msg = (
            "This schedule may be part of a subscription. "
            "Deleting such a schedule will remove the subscription as well. "
            "This action cannot be undone. "
            f"Are you sure you want to delete the schedule '{self.name}'"
            f" with ID: {self.id}?[Y/N]: "
        )
        self._delete_success_msg = (f"Deleted schedule '{self.name}' with ID: {self.id}.")

        return super().delete(force=force)

    @property
    def expired(self):
        """Whether or not the schedule is expired"""
        return self._expired

    @property
    def schedule_next_delivery(self):
        """Next delivery date of time based schedule"""
        return self._schedule_next_delivery



import logging
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union

from packaging import version

from mstrio import config
from mstrio.api import monitors
from mstrio.connection import Connection
from mstrio.server import Cluster
from mstrio.utils import helper
from mstrio.utils.version_helper import class_version_handler

if TYPE_CHECKING:
    from mstrio.users_and_groups.user import User

logger = logging.getLogger(__name__)


@class_version_handler('11.2.0000')
class UserConnections:
    """Browse and manage active user connections on the environment. Use the
    `fetch()` method to fetch the latest active user connections. Filter the
    `user_connections` by using `filter_connections()` method.

    Attributes:
        connection: A MicroStrategy connection object
        user_connections: All active user connections on the environment
    """

    def __init__(self, connection: Connection):
        """Initialize the `UserConnections` object.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`
        """
        self.connection = connection
        self.user_connections: List[Dict[str, Any]] = []

    def fetch(self) -> None:
        """Populate the `UserConnections` object by retrieving all active user
        connections on the environment."""
        self.user_connections = self.list_connections()

    def filter_connections(self, **filters) -> Union[List[Dict[str, Any]], None]:
        """Filter the user connections stored in the `UserConnections` object
        by specifying the `filters` keyword arguments.

        Args:
            **filters: Available filter parameters: ['id', 'parent_id',
                'username', 'user_full_name', 'project_index', 'project_id',
                'project_name', 'open_jobs_count', 'project_type',
                'date_connection_created', 'duration', 'session_id', 'client',
                'config_level']
        """
        filtered_connections = None
        if not self.user_connections:
            helper.exception_handler(
                "Populate the `UserConnections` object with `UserConnections.fetch` first.",
                Warning
            )
        else:
            filtered_connections = helper.filter_list_of_dicts(self.user_connections, **filters)
        return filtered_connections

    def list_connections(
        self, nodes: Union[str, List[str]] = None, limit: Optional[int] = None, **filters
    ) -> List[Dict[str, Any]]:
        """Get all active user connections. Optionally filter the connections
        by specifying the `filters` keyword arguments.

        Args:
            nodes: Node (server) names on which users will be disconnected.
            limit: limit the number of elements returned. If `None`, all objects
                are returned.
            **filters: Available filter parameters: ['id', 'parent_id',
                'username', 'user_full_name', 'project_index', 'project_id',
                'project_name', 'open_jobs_count', 'project_type',
                'date_connection_created', 'duration', 'session_id', 'client',
                'config_level']
        """
        # TODO: This fully initialises a Cluster object every time the function
        # is run. It would be better to somehow cache it for a given connection.
        all_nodes = Cluster(self.connection).list_nodes(to_dictionary=True)
        all_connections = []
        if nodes is None:
            nodes = [node['name'] for node in all_nodes if node['status'] == 'running']
        else:
            nodes = nodes if isinstance(nodes, list) else [nodes]

        msg = 'Error fetching chunk of active user connections.'
        for node in nodes:
            all_connections.extend(
                helper.fetch_objects_async(
                    self.connection,
                    monitors.get_user_connections,
                    monitors.get_user_connections_async,
                    dict_unpack_value="userConnections",
                    limit=limit,
                    chunk_size=1000,
                    error_msg=msg,
                    node_name=node,
                    filters=filters
                )
            )
        return all_connections

    def disconnect_users(
        self,
        connection_ids: Union[str, List[str]] = None,
        users: Optional[Union[List["User"], List[str]]] = None,
        nodes: Union[str, List[str]] = None,
        force: bool = False,
        **filters
    ) -> Union[List[dict], None]:
        """Disconnect user connections by passing in users (objects) or
        connection_ids. Optionally disconnect users by specifying the `filters`
        keyword arguments.

        Args:
            connection_ids: chosen ids that can be retrieved with
                `list_connections()`
            users: List of User objects or usernames
            nodes: Node (server) names on which users will be disconnected
            force: if True, no additional prompt will be shown before
                disconnecting users
            **filters: Available filter parameters: ['id', 'parent_id',
                'username', 'user_full_name', 'project_index', 'project_id',
                'project_name', 'open_jobs_count', 'project_type',
                'date_connection_created', 'duration', 'session_id', 'client',
                'config_level']
        Returns:
            - list of statuses of disconnecting chosen connections with theirs
              ids and messages from the I-Server:
                status code 200 is when all connections were disconnected
                status code 207 is when some connections were disconnected
                status code 403 without error code is when no connections were
                disconnected
            - in case of error of nothing to disconnect it returns None


        """
        from mstrio.users_and_groups.user import User  # import here to avoid circular imports

        if self.connection and not connection_ids and not users and not filters and not force:
            msg = (
                "You need to pass connection_ids or users or specify filters. To disconnect "
                "all connections use `disconnect_all_users()` method."
            )
            helper.exception_handler(msg)

        if connection_ids:
            # disconnect specific user connections without fetching connections
            return self.__disconnect_by_connection_id(connection_ids)
        else:
            # get all user connections to filter locally
            all_connections = self.list_connections(nodes, **filters)
            if users:  # filter user connections by user objects
                users = users if isinstance(users, list) else [users]
                usernames = []
                for user in users:
                    if isinstance(user, User):
                        usernames.append(user.username)
                    elif isinstance(user, str):
                        usernames.append(user)
                    else:
                        helper.exception_handler(
                            "'user' param must be a list of User objects or usernames.",
                            exception_type=TypeError
                        )

                all_connections = list(
                    filter(lambda conn: conn['username'] in usernames, all_connections)
                )

            if all_connections:
                # extract connection ids and disconnect
                connection_ids = [conn['id'] for conn in all_connections]
                return self.__disconnect_by_connection_id(connection_ids)
            elif config.verbose:
                logger.info('No active user connections.')

    def disconnect_all_users(self, force: bool = False) -> Union[List[dict], None]:
        """Disconnect all user connections.

        Args:
            force: if True, no additional prompt will be shown before
                disconnecting all users

        Returns:
            - list of statuses of disconnecting all connections with their ids
               and messages from the I-Server
            - in case of error it returns None
        """
        if not force:
            user_input = input(
                "Are you sure you want to disconnect all users from the I-Server? [Y/N]: "
            )
            if user_input != "Y":
                return None
            else:
                force = True

        return self.disconnect_users(force=force)

    def __disconnect_by_connection_id(
        self, connection_ids: Union[str, List[str]]
    ) -> Union[List[dict], None]:
        """It disconnects connections which ids are provided in
        'connection_ids'. It prints information about executed operations.
        Returns list of statuses of for the given connection ids with the
        messages from the I-Server or `None` in case of an error.
        """
        connection_ids = connection_ids if isinstance(connection_ids, list) else [connection_ids]
        server_version = helper.version_cut(self.connection.iserver_version)

        # use monitors.delete_user_connections
        # or monitors.delete_user_connection depending on the server version
        if version.parse(server_version) >= version.parse('11.3.1'):
            res = monitors.delete_user_connections(connection=self.connection, ids=connection_ids)
            if res.status_code in [200, 207] or (res.status_code == 403
                                                 and not res.json().get('code', None)):
                return self._prepare_disconnect_by_id_message(
                    statuses=res.json()['deleteUserConnectionsStatus']
                )
            else:
                err_msg = f'Error disconnecting user sessions: {connection_ids}.'
                return helper.response_handler(response=res, msg=err_msg, throw_error=False)
        else:
            # TODO: This can probably be made more elegant and potentially
            # performant if we use the ID as dict key and status as value
            # and maybe use comprehension instead of appending in loop.
            statuses: List[Dict[str, Union[str, int]]] = []
            for connection_id in connection_ids:
                response = monitors.delete_user_connection(
                    self.connection, connection_id, bulk=True
                )
                statuses.append({'id': connection_id, 'status': response.status_code})
            return self._prepare_disconnect_by_id_message(statuses=statuses)

    @staticmethod
    def _prepare_disconnect_by_id_message(statuses: List[dict]) -> List[dict]:
        succeeded = []
        failed = []

        for s in statuses:
            status = s['status']
            id = s['id']
            if status >= 200 and status < 300:
                succeeded.append(id)
            elif status >= 400:
                # in most cases there will be status code 403
                failed.append(id)

        if config.verbose:
            if succeeded:
                logger.info(
                    'User connections with ids below were successfully disconnected:\n\t'
                    + ',\n\t'.join(succeeded)
                )
            if failed:
                logger.warning(
                    'User connections with ids below were not disconnected:\n\t'
                    + ',\n\t'.join(failed)
                )

        return statuses



from functools import wraps
from inspect import getattr_static, getmembers

from mstrio.api.exceptions import VersionException
from mstrio.connection import Connection


def method_version_handler(version):
    """
    Decorator which can be applied to a function.
    When applied, validates the iServer version of the existing connection
    (retrieved from kwargs, args or `connection` property from self)
    and compares it with the `version` provided in the arguments -
    version at which the functionality becomes applicable.
    """

    def wrapper(function):

        @wraps(function)
        def inner(*args, **kwargs):
            if conn := kwargs.get("connection"):
                connection_obj = conn
            elif conn := list(filter(lambda arg: isinstance(arg, Connection), args)):
                connection_obj = conn[0]
            else:
                connection_obj = (
                    getattr(args[0], 'connection', None) or getattr(args[0], '_connection')
                )

            if connection_obj._iserver_version < version:
                raise VersionException(
                    f"Environments must run IServer version {version} or newer. "
                    "Please update your environments to use this feature."
                )
            return function(*args, **kwargs)

        return inner

    return wrapper


def class_version_handler(version):
    """
    Decorator which can be used on a class.
    It applies the `method_version_handler decorator` to all the methods
    in the decorated class that are not inherited.
    """
    decorator = method_version_handler(version)

    def wrapper(cls):
        for name, method in getmembers(cls):
            if name in cls.__dict__ and callable(method) and not isinstance(
                    getattr_static(cls, name), staticmethod):
                setattr(cls, name, decorator(method))
        return cls

    return wrapper
